import math
import warnings

import numpy as np
from matplotlib import pyplot as plt

import pypulseq as pp

# ======
# SETUP
# ======
dG = 250e-6

# Set system limits
system = pp.Opts(max_grad=32, grad_unit='mT/m', max_slew=130, slew_unit='T/m/s', rf_ringdown_time=100e-6,
                 rf_dead_time=100e-6, adc_dead_time=10e-6)

seq = pp.Sequence(system)  # Create a new sequence object
# Define FOV and resolution
fov = 256e-3
Nx, Ny = 128, 128
n_echo = 16
n_slices = 1
rf_flip = 180
if isinstance(rf_flip, int):
    rf_flip = np.zeros(n_echo) + rf_flip
slice_thickness = 5e-3
TE = 12e-3
TR = 2000e-3
TE_eff = 60e-3
k0 = round(TE_eff / TE)
pe_type = 'linear'

readout_time = 6.4e-3 + 2 * system.adc_dead_time
t_ex = 2.5e-3
t_exwd = t_ex + system.rf_ringdown_time + system.rf_dead_time
t_ref = 2e-3
t_refwd = t_ref + system.rf_ringdown_time + system.rf_dead_time
t_sp = 0.5 * (TE - readout_time - t_refwd)
t_spex = 0.5 * (TE - t_exwd - t_refwd)
fsp_r = 1
fsp_s = 0.5

rf_ex_phase = np.pi / 2
rf_ref_phase = 0

# ======
# CREATE EVENTS
# ======
flip_ex = 90 * np.pi / 180
rf_ex, gz, _ = pp.make_sinc_pulse(flip_angle=flip_ex, system=system, duration=t_ex, slice_thickness=slice_thickness,
                                  apodization=0.5, time_bw_product=4, phase_offset=rf_ex_phase, return_gz=True)
gs_ex = pp.make_trapezoid(channel='z', system=system, amplitude=gz.amplitude, flat_time=t_exwd, rise_time=dG)

flip_ref = rf_flip[0] * np.pi / 180
rf_ref, gz, _ = pp.make_sinc_pulse(flip_angle=flip_ref, system=system, duration=t_ref, slice_thickness=slice_thickness,
                                   apodization=0.5, time_bw_product=4, phase_offset=rf_ref_phase, use='refocusing',
                                   return_gz=True)
gs_ref = pp.make_trapezoid(channel='z', system=system, amplitude=gs_ex.amplitude, flat_time=t_refwd, rise_time=dG)

ags_ex = gs_ex.area / 2
gs_spr = pp.make_trapezoid(channel='z', system=system, area=ags_ex * (1 + fsp_s), duration=t_sp, rise_time=dG)
gs_spex = pp.make_trapezoid(channel='z', system=system, area=ags_ex * fsp_s, duration=t_spex, rise_time=dG)

delta_k = 1 / fov
k_width = Nx * delta_k

gr_acq = pp.make_trapezoid(channel='x', system=system, flat_area=k_width, flat_time=readout_time, rise_time=dG)
adc = pp.make_adc(num_samples=Nx, duration=gr_acq.flat_time - 40e-6, delay=20e-6)
gr_spr = pp.make_trapezoid(channel='x', system=system, area=gr_acq.area * fsp_r, duration=t_sp, rise_time=dG)
gr_spex = pp.make_trapezoid(channel='x', system=system, area=gr_acq.area * (1 + fsp_r), duration=t_spex, rise_time=dG)

agr_spr = gr_spr.area
agr_preph = gr_acq.area / 2 + agr_spr
gr_preph = pp.make_trapezoid(channel='x', system=system, area=agr_preph, duration=t_spex, rise_time=dG)

# Phase-encoding
n_ex = math.floor(Ny / n_echo)
pe_steps = np.arange(1, n_echo * n_ex + 1) - 0.5 * n_echo * n_ex - 1
if divmod(n_echo, 2)[1] == 0:
    pe_steps = np.roll(pe_steps, -round(n_ex / 2))
pe_order = pe_steps.reshape((n_ex, n_echo), order='F').T
phase_areas = pe_order * delta_k

# Split gradients and recombine into blocks
gs1_times = [0, gs_ex.rise_time]
gs1_amp = [0, gs_ex.amplitude]
gs1 = pp.make_extended_trapezoid(channel='z', times=gs1_times, amplitudes=gs1_amp)

gs2_times = [0, gs_ex.flat_time]
gs2_amp = [gs_ex.amplitude, gs_ex.amplitude]
gs2 = pp.make_extended_trapezoid(channel='z', times=gs2_times, amplitudes=gs2_amp)

gs3_times = [0, gs_spex.rise_time, gs_spex.rise_time + gs_spex.flat_time,
             gs_spex.rise_time + gs_spex.flat_time + gs_spex.fall_time]
gs3_amp = [gs_ex.amplitude, gs_spex.amplitude, gs_spex.amplitude, gs_ref.amplitude]
gs3 = pp.make_extended_trapezoid(channel='z', times=gs3_times, amplitudes=gs3_amp)

gs4_times = [0, gs_ref.flat_time]
gs4_amp = [gs_ref.amplitude, gs_ref.amplitude]
gs4 = pp.make_extended_trapezoid(channel='z', times=gs4_times, amplitudes=gs4_amp)

gs5_times = [0, gs_spr.rise_time, gs_spr.rise_time + gs_spr.flat_time,
             gs_spr.rise_time + gs_spr.flat_time + gs_spr.fall_time]
gs5_amp = [gs_ref.amplitude, gs_spr.amplitude, gs_spr.amplitude, 0]
gs5 = pp.make_extended_trapezoid(channel='z', times=gs5_times, amplitudes=gs5_amp)

gs7_times = [0, gs_spr.rise_time, gs_spr.rise_time + gs_spr.flat_time,
             gs_spr.rise_time + gs_spr.flat_time + gs_spr.fall_time]
gs7_amp = [0, gs_spr.amplitude, gs_spr.amplitude, gs_ref.amplitude]
gs7 = pp.make_extended_trapezoid(channel='z', times=gs7_times, amplitudes=gs7_amp)

# Readout gradient
gr3 = gr_preph

gr5_times = [0, gr_spr.rise_time, gr_spr.rise_time + gr_spr.flat_time,
             gr_spr.rise_time + gr_spr.flat_time + gr_spr.fall_time]
gr5_amp = [0, gr_spr.amplitude, gr_spr.amplitude, gr_acq.amplitude]
gr5 = pp.make_extended_trapezoid(channel='x', times=gr5_times, amplitudes=gr5_amp)

gr6_times = [0, readout_time]
gr6_amp = [gr_acq.amplitude, gr_acq.amplitude]
gr6 = pp.make_extended_trapezoid(channel='x', times=gr6_times, amplitudes=gr6_amp)

gr7_times = [0, gr_spr.rise_time, gr_spr.rise_time + gr_spr.flat_time,
             gr_spr.rise_time + gr_spr.flat_time + gr_spr.fall_time]
gr7_amp = [gr_acq.amplitude, gr_spr.amplitude, gr_spr.amplitude, 0]
gr7 = pp.make_extended_trapezoid(channel='x', times=gr7_times, amplitudes=gr7_amp)

# Fill-times
t_ex = gs1.t[-1] + gs2.t[-1] + gs3.t[-1]
t_ref = gs4.t[-1] + gs5.t[-1] + gs7.t[-1] + readout_time
t_end = gs4.t[-1] + gs5.t[-1]
TE_train = t_ex + n_echo * t_ref + t_end
TR_fill = (TR - n_slices * TE_train) / n_slices
TR_fill = system.grad_raster_time * round(TR_fill / system.grad_raster_time)  # Round to gradient raster
if TR_fill < 0:
    TR_fill = 1e-3
    warnings.warn(f'TR too short, adapted to include all slices to: {1000 * n_slices * (TE_train + TR_fill)} ms')
else:
    print(f'TR fill: {1000 * TR_fill} ms')
delay_TR = pp.make_delay(TR_fill)

# ======
# CONSTRUCT SEQUENCE
# ======
for k_ex in range(n_ex + 1):
    for s in range(n_slices):
        rf_ex.freq_offset = gs_ex.amplitude * slice_thickness * (s - (n_slices - 1) / 2)
        rf_ref.freq_offset = gs_ref.amplitude * slice_thickness * (s - (n_slices - 1) / 2)
        rf_ex.phase_offset = rf_ex_phase - 2 * np.pi * rf_ex.freq_offset * pp.calc_rf_center(rf_ex)[0]
        rf_ref.phase_offset = rf_ref_phase - 2 * np.pi * rf_ref.freq_offset * pp.calc_rf_center(rf_ref)[0]

        seq.add_block(gs1)
        seq.add_block(gs2, rf_ex)
        seq.add_block(gs3, gr3)

        for k_echo in range(n_echo):
            if k_ex > 0:
                phase_area = phase_areas[k_echo, k_ex - 1]
            else:
                phase_area = 0.0  # 0.0 and not 0 because -phase_area should successfully result in negative zero

            gp_pre = pp.make_trapezoid(channel='y', system=system, area=phase_area, duration=t_sp, rise_time=dG)
            gp_rew = pp.make_trapezoid(channel='y', system=system, area=-phase_area, duration=t_sp, rise_time=dG)
            seq.add_block(gs4, rf_ref)
            seq.add_block(gs5, gr5, gp_pre)
            if k_ex > 0:
                seq.add_block(gr6, adc)
            else:
                seq.add_block(gr6)

            seq.add_block(gs7, gr7, gp_rew)

        seq.add_block(gs4)
        seq.add_block(gs5)
        seq.add_block(delay_TR)

# ======
# VISUALIZATION
# ======
seq.plot()

# Plot k-spaces
ktraj_adc, ktraj, t_excitation, t_refocusing, _ = seq.calculate_kspace()
plt.plot(ktraj.T)  # Plot the entire k-space trajectory
plt.figure()
plt.plot(ktraj[0], ktraj[1], 'b')  # 2D plot
plt.axis('equal')  # Enforce aspect ratio for the correct trajectory display
plt.plot(ktraj_adc[0], ktraj_adc[1], 'r.')
plt.show()

seq.write('tse_pypulseq.seq')



"""
Demo low-performance EPI sequence which doesn't use ramp-sampling.
"""
import math

import matplotlib.pyplot as plt
import numpy as np

from pypulseq.Sequence.sequence import Sequence
from pypulseq.make_adc import make_adc
from pypulseq.make_sinc_pulse import make_sinc_pulse
from pypulseq.make_trap_pulse import make_trapezoid
from pypulseq.opts import Opts

# ======
# SETUP
# ======
seq = Sequence()  # Create a new sequence object
# Define FOV and resolution
fov = 220e-3
Nx = 64
Ny = 64
slice_thickness = 3e-3  # Slice thickness
n_slices = 3

# Set system limits
system = Opts(max_grad=32, grad_unit='mT/m', max_slew=130, slew_unit='T/m/s', rf_ringdown_time=30e-6,
              rf_dead_time=100e-6)

# ======
# CREATE EVENTS
# ======
# Create 90 degree slice selection pulse and gradient
rf, gz, _ = make_sinc_pulse(flip_angle=np.pi / 2, system=system, duration=3e-3, slice_thickness=slice_thickness,
                            apodization=0.5, time_bw_product=4, return_gz=True)

# Define other gradients and ADC events
delta_k = 1 / fov
k_width = Nx * delta_k
dwell_time = 4e-6
readout_time = Nx * dwell_time
flat_time = math.ceil(readout_time * 1e5) * 1e-5  # round-up to the gradient raster
gx = make_trapezoid(channel='x', system=system, amplitude=k_width / readout_time, flat_time=flat_time)
adc = make_adc(num_samples=Nx, duration=readout_time,
               delay=gx.rise_time + flat_time / 2 - (readout_time - dwell_time) / 2)

# Pre-phasing gradients
pre_time = 8e-4
gx_pre = make_trapezoid(channel='x', system=system, area=-gx.area / 2, duration=pre_time)
gz_reph = make_trapezoid(channel='z', system=system, area=-gz.area / 2, duration=pre_time)
gy_pre = make_trapezoid(channel='y', system=system, area=-Ny / 2 * delta_k, duration=pre_time)

# Phase blip in shortest possible time
dur = math.ceil(2 * math.sqrt(delta_k / system.max_slew) / 10e-6) * 10e-6
gy = make_trapezoid(channel='y', system=system, area=delta_k, duration=dur)

# ======
# CONSTRUCT SEQUENCE
# ======
# Define sequence blocks
for s in range(n_slices):
    rf.freq_offset = gz.amplitude * slice_thickness * (s - (n_slices - 1) / 2)
    seq.add_block(rf, gz)
    seq.add_block(gx_pre, gy_pre, gz_reph)
    for i in range(Ny):
        seq.add_block(gx, adc)  # Read one line of k-space
        seq.add_block(gy)  # Phase blip
        gx.amplitude = -gx.amplitude  # Reverse polarity of read gradient

# ======
# VISUALIZATION
# ======
seq.plot()  # Plot sequence waveforms

ktraj_adc, ktraj, t_excitation, t_refocusing, t_adc = seq.calculate_kspace()

# Plot k-spaces
time_axis = np.arange(1, ktraj.shape[1] + 1) * system.grad_raster_time
plt.plot(time_axis, ktraj.T)  # Plot the entire k-space trajectory
plt.plot(t_adc, ktraj_adc[0, :], '.')  # Sampling points on the kx-axis
plt.figure()
plt.plot(ktraj[0, :], ktraj[1, :], 'b')  # 2D plot
plt.axis('equal')  # Enforce aspect ratio for the correct trajectory display
plt.plot(ktraj_adc[0, :], ktraj_adc[1, :], 'r.')
plt.show()

seq.write('epi_pypulseq.seq')



import math
from collections import OrderedDict
from types import SimpleNamespace
from typing import Tuple
from typing import Union
from warnings import warn

import matplotlib as mpl
import numpy as np
from matplotlib import pyplot as plt

from pypulseq import major, minor, revision
from pypulseq.Sequence import block
from pypulseq.Sequence import parula
from pypulseq.Sequence.read_seq import read
from pypulseq.Sequence.test_report import test_report as ext_test_report
from pypulseq.Sequence.write_seq import write
from pypulseq.calc_rf_center import calc_rf_center
from pypulseq.check_timing import check_timing as ext_check_timing
from pypulseq.decompress_shape import decompress_shape
from pypulseq.event_lib import EventLibrary
from pypulseq.opts import Opts
from pypulseq.points_to_waveform import points_to_waveform
from pypulseq.supported_labels import get_supported_labels


class Sequence:
    version_major: int = major
    version_minor: int = minor
    version_revision: int = revision
    if isinstance(version_revision, str):
        version_revision = int(version_revision[0])

    def __init__(self, system=Opts()):
        # =========
        # EVENT LIBRARIES
        # =========
        self.adc_library = EventLibrary()  # Library of ADC events
        self.delay_library = EventLibrary()  # Library of delay events
        # Library of extension events. Extension events form single-linked zero-terminated lists
        self.extensions_library = EventLibrary()
        self.grad_library = EventLibrary()  # Library of gradient events
        self.label_inc_library = EventLibrary()  # Library of Label(inc) events (reference from the extensions library)
        self.label_set_library = EventLibrary()  # Library of Label(set) events (reference from the extensions library)
        self.rf_library = EventLibrary()  # Library of RF events
        self.shape_library = EventLibrary()  # Library of compressed shapes
        self.trigger_library = EventLibrary()  # Library of trigger events

        # =========
        # OTHER
        # =========
        self.system = system

        self.rf_raster_time = self.system.rf_raster_time  # RF raster time (system dependent)
        self.grad_raster_time = self.system.grad_raster_time  # Gradient raster time (system dependent)

        self.dict_block_events = OrderedDict()  # Event table
        self.dict_definitions = OrderedDict()  # Optional sequence dict_definitions

        self.arr_block_durations = []  # Cache of block durations
        self.arr_extension_numeric_idx = []  # numeric IDs of the used extensions
        self.arr_extension_string_idx = []  # string IDs of the used extensions

    def __str__(self):
        s = "Sequence:"
        s += "\nshape_library: " + str(self.shape_library)
        s += "\nrf_library: " + str(self.rf_library)
        s += "\ngrad_library: " + str(self.grad_library)
        s += "\nadc_library: " + str(self.adc_library)
        s += "\ndelay_library: " + str(self.delay_library)
        s += "\nextensions_library: " + str(self.extensions_library)  # inserted for trigger support by mveldmann
        s += "\nrf_raster_time: " + str(self.rf_raster_time)
        s += "\ngrad_raster_time: " + str(self.grad_raster_time)
        s += "\ndict_block_events: " + str(len(self.dict_block_events))
        return s

    def add_block(self, *args: SimpleNamespace) -> None:
        """
        Adds event(s) as a block to `Sequence`.

        Parameters
        ----------
        args
            Event or list of events to be added as a block to `Sequence`.
        """
        block.add_block(self, len(self.dict_block_events) + 1, *args)

    def calculate_kspace(self, trajectory_delay: int = 0) -> Tuple[np.array, np.array, np.array, np.array, np.array]:
        """
        Calculates the k-space trajectory of the entire pulse sequence.

        Parameters
        ----------
        trajectory_delay : int, default=0
            Compensation factor in millis to align ADC and gradients in the reconstruction.

        Returns
        -------
        k_traj_adc : numpy.array
            K-space trajectory sampled at `t_adc` timepoints.
        k_traj : numpy.array
            K-space trajectory of the entire pulse sequence.
        t_excitation : numpy.array
            Excitation timepoints.
        t_refocusing : numpy.array
            Refocusing timepoints.
        t_adc : numpy.array
            Sampling timepoints.
        """
        # Initialise the counters and accumulator objects
        count_excitation = 0
        count_refocusing = 0
        count_adc_samples = 0

        # Loop through the blocks to prepare preallocations
        for block_counter in range(len(self.dict_block_events)):
            block = self.get_block(block_counter + 1)
            if hasattr(block, 'rf'):
                if not hasattr(block.rf, 'use') or block.rf.use != 'refocusing':
                    count_excitation += 1
                else:
                    count_refocusing += 1

            if hasattr(block, 'adc'):
                count_adc_samples += int(block.adc.num_samples)

        t_excitation = np.zeros(count_excitation)
        t_refocusing = np.zeros(count_refocusing)
        k_time = np.zeros(count_adc_samples)
        current_duration = 0
        count_excitation = 0
        count_refocusing = 0
        kc_outer = 0
        traj_recon_delay = trajectory_delay

        # Go through the blocks and collect RF and ADC timing data
        for block_counter in range(len(self.dict_block_events)):
            block = self.get_block(block_counter + 1)

            if hasattr(block, 'rf'):
                rf = block.rf
                rf_center, _ = calc_rf_center(rf)
                t = rf.delay + rf_center
                if not hasattr(block.rf, 'use') or block.rf.use != 'refocusing':
                    t_excitation[count_excitation] = current_duration + t
                    count_excitation += 1
                else:
                    t_refocusing[count_refocusing] = current_duration + t
                    count_refocusing += 1

            if hasattr(block, 'adc'):
                _k_time = np.arange(block.adc.num_samples) + 0.5
                _k_time = _k_time * block.adc.dwell + block.adc.delay + current_duration + traj_recon_delay
                k_time[kc_outer:kc_outer + block.adc.num_samples] = _k_time
                kc_outer += block.adc.num_samples
            current_duration += self.arr_block_durations[block_counter]

        # Now calculate the actual k-space trajectory based on the gradient waveforms
        gw = self.gradient_waveforms()
        i_excitation = np.round(t_excitation / self.grad_raster_time)
        i_refocusing = np.round(t_refocusing / self.grad_raster_time)
        i_periods = np.sort([1, *(i_excitation + 1), *(i_refocusing + 1), gw.shape[1] + 1]).astype(np.int)
        # i_periods -= 1  # Python is 0-indexed
        ii_next_excitation = min(len(i_excitation), 1)
        ii_next_refocusing = min(len(i_refocusing), 1)
        k_traj = np.zeros_like(gw)
        k = np.zeros((3, 1))

        for i in range(len(i_periods) - 1):
            i_period_end = i_periods[i + 1] - 1
            k_period = np.concatenate((k, gw[:, i_periods[i] - 1:i_period_end] * self.grad_raster_time), axis=1)
            k_period = np.cumsum(k_period, axis=1)
            k_traj[:, i_periods[i] - 1:i_period_end] = k_period[:, 1:]
            k = k_period[:, -1]

            if ii_next_excitation > 0 and i_excitation[ii_next_excitation - 1] == i_period_end:
                k[:] = 0
                k_traj[:, i_period_end - 1] = np.nan
                ii_next_excitation = min(len(i_excitation), ii_next_excitation + 1)

            if ii_next_refocusing > 0 and i_refocusing[ii_next_refocusing - 1] == i_period_end:
                k = -k
                ii_next_refocusing = min(len(i_refocusing), ii_next_refocusing + 1)

            k = k.reshape((-1, 1))  # To be compatible with np.concatenate

        k_traj_adc = []
        for _k_traj_row in k_traj:
            result = np.interp(xp=np.array(range(1, k_traj.shape[1] + 1)) * self.grad_raster_time,
                               fp=_k_traj_row,
                               x=k_time)
            k_traj_adc.append(result)
        k_traj_adc = np.stack(k_traj_adc)
        t_adc = k_time

        return k_traj_adc, k_traj, t_excitation, t_refocusing, t_adc

    def check_timing(self) -> Tuple[bool, list]:
        """
        Check timing of the events in each block based on grad raster time system limit.

        Returns
        -------
        is_ok : bool
            Boolean flag indicating timing errors.
        error_report : str
            Error report in case of timing errors.
        """
        error_report = []
        is_ok = True
        num_blocks = len(self.dict_block_events)
        total_duration = 0

        for block_counter in range(num_blocks):
            block = self.get_block(block_counter + 1)
            event_names = ['rf', 'gx', 'gy', 'gz', 'adc', 'delay']
            ind = [hasattr(block, attr) for attr in event_names]
            events = [getattr(block, event_names[i]) for i in range(len(event_names)) if ind[i] == 1]
            res, rep, duration = ext_check_timing(self.system, *events)
            is_ok = is_ok and res
            if len(rep) != 0:
                error_report.append(f'Event: {block_counter} - {rep}\n')
            total_duration += duration

        # Check if all the gradients in the last block are ramped down properly
        if len(events) != 0:
            for e in range(len(events)):
                if not isinstance(events[e], list) and events[e].type == 'grad':
                    if events[e].last != 0:
                        error_report.append(
                            f'Event {block_counter} gradients do not ramp to 0 at the end of the sequence')

        self.set_definition('Total duration', total_duration)

        return is_ok, error_report

    def duration(self) -> Tuple[int, int, np.ndarray]:
        """
        Returns the total duration of this sequence, and the total count of blocks and events.

        Returns
        -------
        duration : int
            Duration of this sequence in millis.
        num_blocks : int
            Number of blocks in this sequence.
        event_count : numpy.ndarray
            Number of events in this sequence.
        """

        num_blocks = len(self.dict_block_events)
        event_count = np.zeros(len(self.dict_block_events[1]))
        duration = 0
        for block_counter in range(num_blocks):
            event_count += self.dict_block_events[block_counter + 1] > 0
            duration += self.arr_block_durations[block_counter]

        return duration, num_blocks, event_count

    def flip_grad_axis(self, axis: str) -> None:
        """
        Convenience function to invert all gradients along specified axis.

        Parameters
        ----------
        axis : str
            Gradients to invert or scale. Must be one of 'x', 'y' or 'z'.
        """
        self.mod_grad_axis(axis, modifier=-1)

    def get_block(self, block_index: int) -> SimpleNamespace:
        """
        Retrieves block of events identified by `block_index` from `Sequence`.

        Parameters
        ----------
        block_index : int
            Index of block to be retrieved from `Sequence`.

        Returns
        -------
        SimpleNamespace
            Event identified by `block_index`.
        """
        return block.get_block(self, block_index)

    def get_definition(self, key: str) -> str:
        """
        Retrieves definition identified by `key` from `Sequence`. Returns `None` if no matching definition is found.

        Parameters
        ----------
        key : str
            Key of definition to retrieve.

        Returns
        -------
        str
            Definition identified by `key` if found, else returns ''.
        """
        if key in self.dict_definitions:
            return self.dict_definitions[key]
        else:
            return ''

    def get_extension_type_ID(self, extension_string: str) -> int:
        """
        Get numeric extension ID for `extension_string`. Will automatically create a new ID if unknown.

        Parameters
        ----------
        extension_string : str
            Given string extension ID.

        Returns
        -------
        extension_id : int
            Numeric ID for given string extension ID.

        """
        if extension_string not in self.arr_extension_string_idx:
            if len(self.arr_extension_numeric_idx) == 0:
                extension_id = 1
            else:
                extension_id = 1 + max(self.arr_extension_numeric_idx)

            self.arr_extension_numeric_idx.append(extension_id)
            self.arr_extension_string_idx.append(extension_string)
            assert len(self.arr_extension_numeric_idx) == len(self.arr_extension_string_idx)
        else:
            num = self.arr_extension_string_idx.index(extension_string)
            extension_id = self.arr_extension_numeric_idx[num]

        return extension_id

    def get_extension_type_string(self, extension_id: int) -> str:
        """
        Get string extension ID for `extension_id`.

        Parameters
        ----------
        extension_id : int
            Given numeric extension ID.

        Returns
        -------
        extension_str : str
            String ID for the given numeric extension ID.

        Raises
        ------
        ValueError
            If given numeric extension ID is unknown.
        """
        if extension_id in self.arr_extension_numeric_idx:
            num = self.arr_extension_numeric_idx.index(extension_id)
        else:
            raise ValueError(f'Extension for the given ID - {extension_id} - is unknown.')

        extension_str = self.arr_extension_string_idx[num]
        return extension_str

    def gradient_waveforms(self) -> np.ndarray:
        """
        Decompress the entire gradient waveform. Returns an array of shape `gradient_axes x timepoints`.
        `gradient_axes` is typically 3.

        Returns
        -------
        grad_waveforms : numpy.ndarray
            Decompressed gradient waveform.
        """
        duration, num_blocks, _ = self.duration()

        wave_length = math.ceil(duration / self.grad_raster_time)
        grad_channels = 3
        grad_waveforms = np.zeros((grad_channels, wave_length))
        grad_channels = ['gx', 'gy', 'gz']

        t0 = 0
        t0_n = 0
        for block_counter in range(num_blocks):
            block = self.get_block(block_counter + 1)
            for j in range(len(grad_channels)):
                if hasattr(block, grad_channels[j]):
                    grad = getattr(block, grad_channels[j])
                    if grad.type == 'grad':
                        nt_start = round((grad.delay + grad.t[0]) / self.grad_raster_time)
                        waveform = grad.waveform
                    else:
                        nt_start = round(grad.delay / self.grad_raster_time)
                        if abs(grad.flat_time) > np.finfo(float).eps:
                            t = np.cumsum([0, grad.rise_time, grad.flat_time, grad.fall_time])
                            trap_form = np.multiply([0, 1, 1, 0], grad.amplitude)
                        else:
                            t = np.cumsum([0, grad.rise_time, grad.fall_time])
                            trap_form = np.multiply([0, 1, 0], grad.amplitude)

                        tn = math.floor(t[-1] / self.grad_raster_time)
                        t = np.append(t, t[-1] + self.grad_raster_time)
                        trap_form = np.append(trap_form, 0)

                        if abs(grad.amplitude) > np.finfo(float).eps:
                            waveform = points_to_waveform(times=t, amplitudes=trap_form,
                                                          grad_raster_time=self.grad_raster_time)
                        else:
                            waveform = np.zeros(tn + 1)

                    if len(waveform) != np.sum(np.isfinite(waveform)):
                        warn('Not all elements of the generated waveform are finite')

                    """
                    Matlab dynamically resizes arrays during slice assignment operation if assignment is out of bounds
                    Numpy does not; following is a workaround
                    """
                    l1, l2 = int(t0_n + nt_start), int(t0_n + nt_start + len(waveform))
                    if l2 > grad_waveforms.shape[1]:
                        z = np.zeros((grad_waveforms.shape[0], l2 - grad_waveforms.shape[1]))
                        grad_waveforms = np.hstack((grad_waveforms, z))
                    grad_waveforms[j, l1:l2] = waveform

            t0 += self.arr_block_durations[block_counter]
            t0_n = round(t0 / self.grad_raster_time)

        return grad_waveforms

    def mod_grad_axis(self, axis: str, modifier: int) -> None:
        """
        Invert or scale all gradients along the corresponding axis/channel. The function acts on all gradient objects
        already added to the sequence object.

        Parameters
        ----------
        axis : str
            Gradients to invert or scale. Must be one of 'x', 'y' or 'z'.
        modifier : int
            Scaling value.

        Raises
        ------
        ValueError
            If invalid `axis` is passed. Must be one of 'x', 'y','z'.
        RuntimeError
            If same gradient event is used on multiple axes.
        """
        if axis not in ['x', 'y', 'z']:
            raise ValueError(f"Invalid axis. Must be one of 'x', 'y','z'. Passed: {axis}")

        channel_num = ['x', 'y', 'z'].index(axis)
        other_channels = [0, 1, 2]
        other_channels.remove(channel_num)

        # Go through all event table entries and list gradient objects in the library
        all_grad_events = np.array(list(self.dict_block_events.values()))
        all_grad_events = all_grad_events[:, 2:5]

        selected_events = np.unique(all_grad_events[:, channel_num])
        selected_events = selected_events[selected_events != 0]
        other_events = np.unique(all_grad_events[:, other_channels])
        if len(np.intersect1d(selected_events, other_events)) > 0:
            raise RuntimeError('mod_grad_axis does not yet support the same gradient event used on multiple axes.')

        for i in range(len(selected_events)):
            self.grad_library.data[selected_events[i]][0] *= modifier
            if self.grad_library.type[selected_events[i]] == 'g' and self.grad_library.lengths[selected_events[i]] == 5:
                # Need to update first and last fields
                self.grad_library.data[selected_events[i]][3] *= modifier
                self.grad_library.data[selected_events[i]][4] *= modifier

    def plot(self, label: str = str(), save: bool = False, time_range=(0, np.inf), time_disp: str = 's',
             plot_type: str = 'Gradient') -> None:
        """
        Plot `Sequence`.

        Parameters
        ----------
        label : str, defualt=str()

        save : bool, default=False
            Boolean flag indicating if plots should be saved. The two figures will be saved as JPG with numerical
            suffixes to the filename 'seq_plot'.
        time_range : iterable, default=(0, np.inf)
            Time range (x-axis limits) for plotting the sequence. Default is 0 to infinity (entire sequence).
        time_disp : str, default='s'
            Time display type, must be one of `s`, `ms` or `us`.
        plot_type : str, default='Gradient'
            Gradients display type, must be one of either 'Gradient' or 'Kspace'.
        """
        mpl.rcParams['lines.linewidth'] = 0.75  # Set default Matplotlib linewidth

        valid_plot_types = ['Gradient', 'Kspace']
        valid_time_units = ['s', 'ms', 'us']
        valid_labels = get_supported_labels()
        if plot_type not in valid_plot_types:
            raise ValueError('Unsupported plot type')
        if not all([isinstance(x, (int, float)) for x in time_range]) or len(time_range) != 2:
            raise ValueError('Invalid time range')
        if time_disp not in valid_time_units:
            raise ValueError('Unsupported time unit')

        fig1, fig2 = plt.figure(1), plt.figure(2)
        sp11 = fig1.add_subplot(311)
        sp12, sp13 = fig1.add_subplot(312, sharex=sp11), fig1.add_subplot(313, sharex=sp11)
        fig2_subplots = [fig2.add_subplot(311, sharex=sp11), fig2.add_subplot(312, sharex=sp11),
                         fig2.add_subplot(313, sharex=sp11)]

        t_factor_list = [1, 1e3, 1e6]
        t_factor = t_factor_list[valid_time_units.index(time_disp)]

        t0 = 0
        label_defined = False
        label_idx_to_plot = []
        label_legend_to_plot = []
        label_store = dict()
        for i in range(len(valid_labels)):
            label_store[valid_labels[i]] = 0
            if label.upper() == valid_labels[i]:
                label_idx_to_plot.append(i)
                label_legend_to_plot.append(valid_labels[i])

        if len(label_idx_to_plot) != 0:
            p = parula.main(len(label_idx_to_plot) + 1)
            label_colors_to_plot = p(np.arange(len(label_idx_to_plot)))

        for block_counter in range(len(self.dict_block_events)):
            block = self.get_block(block_counter + 1)
            is_valid = time_range[0] <= t0 <= time_range[1]
            if is_valid:
                if hasattr(block, 'label'):
                    for i in range(len(block.label)):
                        if block.label[i].type == 'labelinc':
                            label_store[block.label[i].label] += block.label[i].value
                        else:
                            label_store[block.label[i].label] = block.label[i].value
                    label_defined = True

                if hasattr(block, 'adc'):
                    adc = block.adc
                    # From Pulseq: According to the information from Klaus Scheffler and indirectly from Siemens this
                    # is the present convention - the samples are shifted by 0.5 dwell
                    t = adc.delay + (np.arange(int(adc.num_samples)) + 0.5) * adc.dwell
                    sp11.plot(t_factor * (t0 + t), np.zeros(len(t)), 'rx')
                    sp13.plot(t_factor * (t0 + t),
                              np.angle(np.exp(1j * adc.phase_offset) * np.exp(1j * 2 * np.pi * t * adc.freq_offset)),
                              'b.')

                    if label_defined and len(label_idx_to_plot) != 0:
                        cycler = mpl.cycler(color=label_colors_to_plot)
                        sp11.set_prop_cycle(cycler)
                        label_store_arr = list(label_store.values())
                        lbl_vals = np.take(label_store_arr, label_idx_to_plot)
                        t = t0 + adc.delay + (adc.num_samples - 1) / 2 * adc.dwell
                        p = sp11.plot(t_factor * t, lbl_vals, '.')
                        if len(label_legend_to_plot) != 0:
                            sp11.legend(p, label_legend_to_plot, loc='upper left')
                            label_legend_to_plot = []

                if hasattr(block, 'rf'):
                    rf = block.rf
                    tc, ic = calc_rf_center(rf)
                    t = rf.t + rf.delay
                    tc = tc + rf.delay
                    sp12.plot(t_factor * (t0 + t), np.abs(rf.signal))
                    sp13.plot(t_factor * (t0 + t), np.angle(rf.signal * np.exp(1j * rf.phase_offset)
                                                            * np.exp(1j * 2 * math.pi * rf.t * rf.freq_offset)),
                              t_factor * (t0 + tc), np.angle(rf.signal[ic] * np.exp(1j * rf.phase_offset)
                                                             * np.exp(1j * 2 * math.pi * rf.t[ic] * rf.freq_offset)),
                              'xb')

                grad_channels = ['gx', 'gy', 'gz']
                for x in range(len(grad_channels)):
                    if hasattr(block, grad_channels[x]):
                        grad = getattr(block, grad_channels[x])
                        if grad.type == 'grad':
                            # In place unpacking of grad.t with the starred expression
                            t = grad.delay + [0, *(grad.t + (grad.t[1] - grad.t[0]) / 2),
                                              grad.t[-1] + grad.t[1] - grad.t[0]]
                            waveform = 1e-3 * np.array((grad.first, *grad.waveform, grad.last))
                        else:
                            t = np.cumsum([0, grad.delay, grad.rise_time, grad.flat_time, grad.fall_time])
                            waveform = 1e-3 * grad.amplitude * np.array([0, 0, 1, 1, 0])
                        fig2_subplots[x].plot(t_factor * (t0 + t), waveform)
            t0 += self.arr_block_durations[block_counter]

        grad_plot_labels = ['x', 'y', 'z']
        sp11.set_ylabel('ADC')
        sp12.set_ylabel('RF mag (Hz)')
        sp13.set_ylabel('RF/ADC phase (rad)')
        sp13.set_xlabel('t(s)')
        for x in range(3):
            _label = grad_plot_labels[x]
            fig2_subplots[x].set_ylabel(f'G{_label} (kHz/m)')
        fig2_subplots[-1].set_xlabel('t(s)')

        # Setting display limits
        disp_range = t_factor * np.array([time_range[0], min(t0, time_range[1])])
        [x.set_xlim(disp_range) for x in [sp11, sp12, sp13, *fig2_subplots]]

        fig1.tight_layout()
        fig2.tight_layout()
        if save:
            fig1.savefig('seq_plot1.jpg')
            fig2.savefig('seq_plot2.jpg')
        plt.show()

    def read(self, file_path: str) -> None:
        """
        Read `.seq` file from `file_path`.

        Parameters
        ----------
        file_path : str
            Path to `.seq` file to be read.
        """
        read(self, file_path)

    def rf_from_lib_data(self, lib_data: list) -> SimpleNamespace:
        """
        Construct RF object from `lib_data`.

        Parameters
        ----------
        lib_data : list
            RF envelope.

        Returns
        -------
        rf : SimpleNamespace
            RF object constructed from lib_data.
        """
        rf = SimpleNamespace()
        rf.type = 'rf'

        amplitude, mag_shape, phase_shape = lib_data[0], lib_data[1], lib_data[2]
        shape_data = self.shape_library.data[mag_shape]
        compressed = SimpleNamespace()
        compressed.num_samples = shape_data[0]
        compressed.data = shape_data[1:]
        mag = decompress_shape(compressed)
        shape_data = self.shape_library.data[phase_shape]
        compressed.num_samples = shape_data[0]
        compressed.data = shape_data[1:]
        phase = decompress_shape(compressed)
        rf.signal = amplitude * mag * np.exp(1j * 2 * np.pi * phase)
        rf.t = np.arange(1, len(mag) + 1) * self.rf_raster_time

        rf.delay = lib_data[3]
        rf.freq_offset = lib_data[4]
        rf.phase_offset = lib_data[5]

        if len(lib_data) < 7:
            lib_data = np.append(lib_data, 0)
        rf.dead_time = lib_data[6]

        if len(lib_data.shape) < 8:
            lib_data = np.append(lib_data, 0)
        rf.ringdown_time = lib_data[7]

        if len(lib_data.shape) < 9:
            lib_data = np.append(lib_data, 0)

        use_cases = {1: 'excitation', 2: 'refocusing', 3: 'inversion'}
        if lib_data[8] in use_cases:
            rf.use = use_cases[lib_data[8]]

        return rf

    def set_definition(self, key: str, val: Union[int, list, np.ndarray, str, tuple]) -> None:
        """
        Sets custom definition to the `Sequence`.

        Parameters
        ----------
        key : str
            Definition key.
        val : int, list, numpy.ndarray, str or tuple
            Definition value.
        """
        if key == 'FOV':
            if max(val) > 1:
                text = 'Definition FOV uses values exceeding 1 m.'
                text += 'New Pulseq interpreters expect values in units of meters.'
                warn(text)

        self.dict_definitions[key] = val

    def set_extension_string_ID(self, extension_str: str, extension_id: int) -> None:
        """
        Set numeric ID for the given string extension ID.

        Parameters
        ----------
        extension_str : str
            Given string extension ID.
        extension_id : int
            Given numeric extension ID.

        Raises
        ------
        ValueError
            If given numeric or string extension ID is not unique.
        """
        if extension_str in self.arr_extension_string_idx or extension_id in self.arr_extension_numeric_idx:
            raise ValueError('Numeric or string ID is not unique')

        self.arr_extension_numeric_idx.append(extension_id)
        self.arr_extension_string_idx.append(extension_str)
        assert len(self.arr_extension_numeric_idx) == len(self.arr_extension_string_idx)

    def test_report(self) -> str:
        """
        Analyze the sequence and return a text report.
        """
        return ext_test_report(self)

    def write(self, name: str) -> None:
        """
        Writes the calling `Sequence` object as a `.seq` file with filename `name`.

        Parameters
        ----------
        name :str
            Filename of `.seq` file to be written to disk.
        """
        write(self, name)



from types import SimpleNamespace
from typing import List, Union

import numpy as np

from pypulseq.calc_duration import calc_duration


def align(**kwargs: Union[SimpleNamespace, List[SimpleNamespace]]) -> List[SimpleNamespace]:
    """
    Aligns `SimpleNamespace` objects as per specified alignment options by setting delays of the pulse sequence events
    within the block. All previously configured delays within objects are taken into account during calculating of the
    block duration but then reset according to the selected alignment. Possible values for align_spec are 'left',
    'center', 'right'.

    Parameters
    ----------
    args : dict[str, list[SimpleNamespace]
        Dictionary mapping of alignment options and `SimpleNamespace` objects.
        Template: alignment_spec1=SimpleNamespace, alignment_spec2=[SimpleNamespace, ...], ...
        Alignment spec must be one of `left`, `center` or `right`.

    Returns
    -------
    objects : list
        List of aligned `SimpleNamespace` objects.

    Raises
    ------
    ValueError
        If first parameter is not of type `str`.
        If invalid alignment spec is passed. Must be one of `left`, `center` or `right`.
    """
    alignment_specs = list(kwargs.keys())
    if not isinstance(alignment_specs[0], str):
        raise ValueError(f'First parameter must be of type str. Passed: {type(alignment_specs[0])}')

    alignment_options = ['left', 'center', 'right']
    if np.any([align_opt not in alignment_options for align_opt in alignment_specs]):
        raise ValueError('Invalid alignment spec.')

    alignments = []
    objects = []
    for a in alignment_specs:
        objects_to_align = kwargs[a]
        a = alignment_options.index(a)
        if isinstance(objects_to_align, (list, np.ndarray, tuple)):
            alignments.extend([a] * len(objects_to_align))
            objects.extend(objects_to_align)
        elif isinstance(objects_to_align, SimpleNamespace):
            alignments.extend([a])
            objects.append(objects_to_align)

    dur = calc_duration(*objects)

    for i in range(len(objects)):
        if alignments[i] == 0:
            objects[i].delay = 0
        elif alignments[i] == 1:
            objects[i].delay = (dur - calc_duration(objects[i])) / 2
        elif alignments[i] == 2:
            objects[i].delay = dur - calc_duration(objects[i]) + objects[i].delay

    return objects



import re
from pathlib import Path
from typing import Dict, Tuple

import numpy as np

from pypulseq.calc_duration import calc_duration
from pypulseq.event_lib import EventLibrary
from pypulseq.supported_labels import get_supported_labels


def read(self, path: str, detect_rf_use: bool = False) -> None:
    """
    Reads a `.seq` file from `path`.

    Parameters
    ----------
    path : Path
        Path of .seq file to be read.
    detect_rf_use : bool, default=False

    Raises
    ------
    ValueError

    RuntimeError
    """

    input_file = open(path, 'r')
    self.shape_library = EventLibrary()
    self.adc_library = EventLibrary()
    self.delay_library = EventLibrary()
    self.grad_library = EventLibrary()
    self.grad_raster_time = self.system.grad_raster_time
    self.rf_library = EventLibrary()
    self.rf_raster_time = self.system.rf_raster_time
    self.label_inc_library = EventLibrary()
    self.label_set_library = EventLibrary()
    self.trigger_library = EventLibrary()

    self.dict_block_events = {}
    self.dict_definitions = {}

    jemris_generated = False

    while True:
        section = __skip_comments(input_file)
        if section == -1:
            break
        if section == '[DEFINITIONS]':
            self.dict_definitions = __read_definitions(input_file)
        elif section == '[JEMRIS]':
            jemris_generated = True
        elif section == '[VERSION]':
            version_major, version_minor, version_revision = __read_version(input_file)

            if version_major != self.version_major:
                raise RuntimeError(f'Unsupported version_major: {version_major}. Expected: {self.version_major}')

            if version_major == 1 and version_minor == 2 and self.version_major == 1 and self.version_minor == 3:
                compatibility_mode_12x_13x = True
            else:
                compatibility_mode_12x_13x = False

                if version_minor != self.version_minor:
                    raise RuntimeError(f'Unsupported version_minor: {version_minor}. Expected: {self.version_minor}')

                if version_revision > self.version_revision:
                    raise RuntimeError(
                        f'Unsupported version_revision: {version_revision}. Expected: {self.version_revision}')

            if not compatibility_mode_12x_13x:
                self.version_major = version_major
                self.version_minor = version_minor
                self.version_revision = version_revision

        elif section == '[BLOCKS]':
            self.dict_block_events = __read_blocks(input_file, compatibility_mode_12x_13x)
        elif section == '[RF]':
            if jemris_generated:
                self.rf_library = __read_events(input_file, (1, 1, 1, 1, 1), event_library=self.rf_library)
            else:
                self.rf_library = __read_events(input_file, (1, 1, 1, 1e-6, 1, 1), event_library=self.rf_library)
        elif section == '[GRADIENTS]':
            self.grad_library = __read_events(input_file, (1, 1, 1e-6), 'g', self.grad_library)
        elif section == '[TRAP]':
            if jemris_generated:
                self.grad_library = __read_events(input_file, (1, 1e-6, 1e-6, 1e-6), 't', self.grad_library)
            else:
                self.grad_library = __read_events(input_file, (1, 1e-6, 1e-6, 1e-6, 1e-6), 't', self.grad_library)
        elif section == '[ADC]':
            self.adc_library = __read_events(input_file, (1, 1e-9, 1e-6, 1, 1), event_library=self.adc_library)
        elif section == '[DELAYS]':
            self.delay_library = __read_events(input_file, (1e-6,), event_library=self.delay_library)
        elif section == '[SHAPES]':
            self.shape_library = __read_shapes(input_file)
        elif section == '[EXTENSIONS]':
            self.extensions_library = __read_events(input_file)
        elif section[:18] == 'extension TRIGGERS':
            extension_id = int(section[18:])
            self.set_extension_string_ID('TRIGGERS', extension_id)
            self.trigger_library = __read_events(input_file, (1, 1, 1e-6, 1e-6), event_library=self.trigger_library)
        elif section[:18] == 'extension LABELSET':
            extension_id = int(section[18:])
            self.set_extension_string_ID('LABELSET', extension_id)
            l1 = lambda s: int(s)
            l2 = lambda s: get_supported_labels().index(s) + 1
            self.label_set_library = __read_and_parse_events(input_file, l1, l2)
        elif section[:18] == 'extension LABELINC':
            extension_id = int(section[18:])
            self.set_extension_string_ID('LABELINC', extension_id)
            l1 = lambda s: int(s)
            l2 = lambda s: get_supported_labels().index(s) + 1
            self.label_inc_library = __read_and_parse_events(input_file, l1, l2)
        else:
            raise ValueError(f'Unknown section code: {section}')

    self.arr_block_durations = np.zeros(len(self.dict_block_events))
    grad_channels = ['gx', 'gy', 'gz']
    grad_prev_last = np.zeros(len(grad_channels))
    for block_counter in range(len(self.dict_block_events)):
        block = self.get_block(block_counter + 1)
        block_duration = calc_duration(block)
        self.arr_block_durations[block_counter] = block_duration
        # We also need to keep track of the event IDs because some PyPulseq files written by external software may contain
        # repeated entries so searching by content will fail
        event_idx = self.dict_block_events[block_counter + 1]
        # Update the objects by filling in the fields not contained in the PyPulseq file
        for j in range(len(grad_channels)):
            if hasattr(block, grad_channels[j]):
                grad = getattr(block, grad_channels[j])
            else:
                grad_prev_last[j] = 0
                continue

            if grad.type == 'grad':
                if grad.delay > 0:
                    grad_prev_last[j] = 0

                if hasattr(grad, 'first'):
                    continue

                grad.first = grad_prev_last[j]
                # Restore samples on the edges of the gradient raster intervals for that we need the first sample
                odd_step1 = [grad.first, *2 * grad.waveform]
                odd_step2 = odd_step1 * (np.mod(range(len(odd_step1)), 2) * 2 - 1)
                waveform_odd_rest = np.cumsum(odd_step2) * (np.mod(len(odd_step2), 2) * 2 - 1)
                grad.lsat = waveform_odd_rest[-1]
                grad_prev_last[j] = grad.last

                eps = np.finfo(np.float).eps
                if grad.delay + len(grad.waveform) * self.grad_raster_time + eps < block_duration:
                    grad_prev_last[j] = 0

                amplitude = np.max(np.abs(grad.waveform))
                old_data = [amplitude, grad.shape_id, grad.delay]
                new_data = [amplitude, grad.shape_id, grad.delay, grad.first, grad.last]
                event_id = event_idx[j + 2]
                # update_data()
            else:
                grad_prev_last[j] = 0

    if detect_rf_use:
        for k in self.rf_library.keys():
            lib_data = self.rf_library.data[k]
            rf = self.rf_from_lib_data(lib_data)
            flip_deg = np.abs(np.sum(rf.signal)) * rf.t[0] * 360
            if len(lib_data) < 9:
                if flip_deg < 90.01:
                    lib_data[8] = 0
                else:
                    lib_data[8] = 2
                self.rf_library.data[k] = lib_data


def __read_definitions(input_file) -> Dict[str, str]:
    """
    Read dict_definitions from .seq file.

    Parameters
    ----------
    input_file : file object
        .seq file

    Returns
    -------
    dict_definitions : dict
        Dict object containing key value pairs of dict_definitions.
    """
    definitions = dict()
    line = __strip_line(input_file)
    while line != '' and line[0] != '#':
        tok = line.split(' ')
        try:  # Try converting every element into a float
            [float(x) for x in tok[1:]]
            definitions[tok[0]] = np.array(tok[1:], dtype=float)
        except ValueError:  # Try clause did not work!
            definitions[tok[0]] = tok[1:]
        line = __strip_line(input_file)

    return definitions


def __read_version(input_file) -> Tuple[int, int, int]:
    """
    Read version from .seq file.

    Parameters
    ----------
    input_file : file object
        .seq file

    Returns
    -------
    tuple
        Tuple of major, minor and revision number.
    """
    line = __strip_line(input_file)
    major, minor, revision = 0, 0, 0
    while line != '' and line[0] != '#':
        tok = line.split(' ')
        if tok[0] == 'major':
            major = int(tok[1])
        elif tok[0] == 'minor':
            minor = int(tok[1])
        elif tok[0] == 'revision':
            revision = int(tok[1])
        else:
            raise RuntimeError()
        line = __strip_line(input_file)

    return major, minor, revision


def __read_blocks(input_file, compatibility_mode_12x_13x: bool) -> dict:
    """
    Read Pulseq blocks from .seq file.

    Parameters
    ----------
    input_file : file
        .seq file
    compatibility_mode_12x_13x : bool

    Returns
    -------
    event_table : dict
        Dict object containing key value pairs of Pulseq block ID and block definition.
    """

    line = __strip_line(input_file)

    event_table = dict()
    while line != '' and line != '#':
        block_events = np.fromstring(line, dtype=int, sep=' ')

        if compatibility_mode_12x_13x:
            event_table[block_events[0]] = np.array([*block_events[1:], 0])
        else:
            event_table[block_events[0]] = block_events[1:]

        line = __strip_line(input_file)

    return event_table


def __read_events(input_file, scale: list = (1,), event_type: str = str(),
                  event_library: EventLibrary = EventLibrary()) -> EventLibrary:
    """
    Read Pulseq events from .seq file.

    Parameters
    ----------
    input_file : file object
        .seq file
    scale : list, default=(1,)
        Scaling factor.
    event_type : str
        Type of Pulseq event.
    event_library : EventLibrary, default=EventLibrary()
        EventLibrary

    Returns
    -------
    dict_definitions : dict
        `EventLibrary` object containing Pulseq event dict_definitions.
    """
    line = __strip_line(input_file)

    while line != '' and line != '#':
        data = np.fromstring(line, dtype=float, sep=' ')
        event_id = data[0]
        data = data[1:] * scale
        if event_type == '':
            event_library.insert(key_id=event_id, new_data=data)
        else:
            event_library.insert(key_id=event_id, new_data=data, data_type=event_type)
        line = __strip_line(input_file)

    return event_library


def __read_and_parse_events(input_file, *args) -> EventLibrary:
    event_library = EventLibrary()
    line = __strip_line(input_file)

    while line != '' and line != '#':
        datas = re.split('(\s+)', line)
        datas = [d for d in datas if d != ' ']
        data = np.zeros(len(datas) - 1, dtype=np.int)
        event_id = int(datas[0])
        for i in range(1, len(datas)):
            if i > len(args):
                data[i - 1] = int(datas[i])
            else:
                data[i - 1] = args[i - 1](datas[i])
        event_library.insert(key_id=event_id, new_data=data)
        line = __strip_line(input_file)

    return event_library


def __read_shapes(input_file) -> EventLibrary:
    """
    Read Pulseq shapes from .seq file.

    Parameters
    ----------
    input_file : file
        .seq file

    Returns
    -------
    shape_library : EventLibrary
        `EventLibrary` object containing shape dict_definitions.
    """
    shape_library = EventLibrary()

    line = __skip_comments(input_file)

    while line != -1 and (line != '' or line[0:8] == 'shape_id'):
        tok = line.split(' ')
        id = int(tok[1])
        line = __skip_comments(input_file)
        tok = line.split(' ')
        num_samples = int(tok[1])
        data = []
        line = __skip_comments(input_file)
        while line != '' and line != '#':
            data.append(float(line))
            line = __strip_line(input_file)
        line = __skip_comments(input_file)
        data.insert(0, num_samples)
        data = np.asarray(data)
        shape_library.insert(key_id=id, new_data=data)
    return shape_library


def __skip_comments(input_file) -> str:
    """
    Skip one '#' comment in .seq file.

    Parameters
    ----------
    input_file : file
        .seq file

    Returns
    -------
    line : str
        First line in `input_file` after skipping one '#' comment block. Note: File pointer is remembered, so successive calls work as expected.
    """

    line = __strip_line(input_file)

    while line != -1 and (line == '' or line[0] == '#'):
        line = __strip_line(input_file)

    return line


def __strip_line(input_file) -> str:
    """
    Removes spaces and newline whitespaces.

    Parameters
    ----------
    input_file : file
        .seq file

    Returns
    -------
    line : str
        First line in input_file after removing spaces and newline whitespaces. Note: File pointer is remembered,
        so successive calls work as expected. Returns -1 for eof.
    """
    line = input_file.readline()  # If line is an empty string, end of the file has been reached
    return line.strip() if line != '' else -1



import math
import warnings

import matplotlib.pyplot as plt
import numpy as np

from pypulseq.Sequence.sequence import Sequence
from pypulseq.calc_rf_center import calc_rf_center
from pypulseq.make_adc import make_adc
from pypulseq.make_delay import make_delay
from pypulseq.make_extended_trapezoid import make_extended_trapezoid
from pypulseq.make_sinc_pulse import make_sinc_pulse
from pypulseq.make_trap_pulse import make_trapezoid
from pypulseq.opts import Opts

# ======
# SETUP
# ======
dG = 250e-6

# Set system limits
system = Opts(max_grad=30, grad_unit='mT/m', max_slew=170, slew_unit='T/m/s', rf_ringdown_time=100e-6,
              rf_dead_time=100e-6, adc_dead_time=10e-6)

seq = Sequence(system=system)  # Create a new sequence object
# Define FOV and resolution
fov = 256e-3
Ny_pre = 8
Nx, Ny = 128, 128
n_echo = int(Ny / 2 + Ny_pre)
n_slices = 1
rf_flip = 180
if isinstance(rf_flip, int):
    rf_flip = np.zeros(n_echo) + rf_flip
slice_thickness = 5e-3  # Slice thickness
TE = 12e-3
TR = 2000e-3
TE_eff = 60e-3
k0 = round(TE_eff / TE)
PE_type = 'linear'

readout_time = 6.4e-3 + 2 * system.adc_dead_time
t_ex = 2.5e-3
t_ex_wd = t_ex + system.rf_ringdown_time + system.rf_dead_time
t_ref = 2e-3
tf_ref_wd = t_ref + system.rf_ringdown_time + system.rf_dead_time
t_sp = 0.5 * (TE - readout_time - tf_ref_wd)
t_sp_ex = 0.5 * (TE - t_ex_wd - tf_ref_wd)
fspR = 1.0
fspS = 0.5

rfex_phase = math.pi / 2
rfref_phase = 0

# ======
# CREATE EVENTS
# ======
# Create 90 degree slice selection pulse and gradient
flipex = 90 * math.pi / 180
rfex, gz, _ = make_sinc_pulse(flip_angle=flipex, system=system, duration=t_ex, slice_thickness=slice_thickness,
                              apodization=0.5, time_bw_product=4, phase_offset=rfex_phase, return_gz=True)
GS_ex = make_trapezoid(channel='z', system=system, amplitude=gz.amplitude, flat_time=t_ex_wd, rise_time=dG)

flipref = rf_flip[0] * math.pi / 180
rfref, gz, _ = make_sinc_pulse(flip_angle=flipref, system=system, duration=t_ref, slice_thickness=slice_thickness,
                               apodization=0.5, time_bw_product=4, phase_offset=rfref_phase, use='refocusing',
                               return_gz=True)
GS_ref = make_trapezoid(channel='z', system=system, amplitude=GS_ex.amplitude, flat_time=tf_ref_wd, rise_time=dG)

AGS_ex = GS_ex.area / 2
GS_spr = make_trapezoid(channel='z', system=system, area=AGS_ex * (1 + fspS), duration=t_sp, rise_time=dG)
GS_spex = make_trapezoid(channel='z', system=system, area=AGS_ex * fspS, duration=t_sp_ex, rise_time=dG)

delta_k = 1 / fov
k_width = Nx * delta_k

GR_acq = make_trapezoid(channel='x', system=system, flat_area=k_width, flat_time=readout_time, rise_time=dG)
adc = make_adc(num_samples=Nx, duration=GR_acq.flat_time - 40e-6, delay=20e-6)
GR_spr = make_trapezoid(channel='x', system=system, area=GR_acq.area * fspR, duration=t_sp, rise_time=dG)
GR_spex = make_trapezoid(channel='x', system=system, area=GR_acq.area * (1 + fspR), duration=t_sp_ex, rise_time=dG)

AGR_spr = GR_spr.area
AGR_preph = GR_acq.area / 2 + AGR_spr
GR_preph = make_trapezoid(channel='x', system=system, area=AGR_preph, duration=t_sp_ex, rise_time=dG)

n_ex = 1
PE_order = np.arange(-Ny_pre, Ny + 1).T
phase_areas = PE_order * delta_k

# Split gradients and recombine into blocks
GS1_times = [0, GS_ex.rise_time]
GS1_amp = [0, GS_ex.amplitude]
GS1 = make_extended_trapezoid(channel='z', times=GS1_times, amplitudes=GS1_amp)

GS2_times = [0, GS_ex.flat_time]
GS2_amp = [GS_ex.amplitude, GS_ex.amplitude]
GS2 = make_extended_trapezoid(channel='z', times=GS2_times, amplitudes=GS2_amp)

GS3_times = [0, GS_spex.rise_time, GS_spex.rise_time + GS_spex.flat_time,
             GS_spex.rise_time + GS_spex.flat_time + GS_spex.fall_time]
GS3_amp = [GS_ex.amplitude, GS_spex.amplitude, GS_spex.amplitude, GS_ref.amplitude]
GS3 = make_extended_trapezoid(channel='z', times=GS3_times, amplitudes=GS3_amp)

GS4_times = [0, GS_ref.flat_time]
GS4_amp = [GS_ref.amplitude, GS_ref.amplitude]
GS4 = make_extended_trapezoid(channel='z', times=GS4_times, amplitudes=GS4_amp)

GS5_times = [0, GS_spr.rise_time, GS_spr.rise_time + GS_spr.flat_time,
             GS_spr.rise_time + GS_spr.flat_time + GS_spr.fall_time]
GS5_amp = [GS_ref.amplitude, GS_spr.amplitude, GS_spr.amplitude, 0]
GS5 = make_extended_trapezoid(channel='z', times=GS5_times, amplitudes=GS5_amp)

GS7_times = [0, GS_spr.rise_time, GS_spr.rise_time + GS_spr.flat_time,
             GS_spr.rise_time + GS_spr.flat_time + GS_spr.fall_time]
GS7_amp = [0, GS_spr.amplitude, GS_spr.amplitude, GS_ref.amplitude]
GS7 = make_extended_trapezoid(channel='z', times=GS7_times, amplitudes=GS7_amp)

# Readout gradient
GR3 = GR_preph

GR5_times = [0, GR_spr.rise_time, GR_spr.rise_time + GR_spr.flat_time,
             GR_spr.rise_time + GR_spr.flat_time + GR_spr.fall_time]
GR5_amp = [0, GR_spr.amplitude, GR_spr.amplitude, GR_acq.amplitude]
GR5 = make_extended_trapezoid(channel='x', times=GR5_times, amplitudes=GR5_amp)

GR6_times = [0, readout_time]
GR6_amp = [GR_acq.amplitude, GR_acq.amplitude]
GR6 = make_extended_trapezoid(channel='x', times=GR6_times, amplitudes=GR6_amp)

GR7_times = [0, GR_spr.rise_time, GR_spr.rise_time + GR_spr.flat_time,
             GR_spr.rise_time + GR_spr.flat_time + GR_spr.fall_time]
GR7_amp = [GR_acq.amplitude, GR_spr.amplitude, GR_spr.amplitude, 0]
GR7 = make_extended_trapezoid(channel='x', times=GR7_times, amplitudes=GR7_amp)

# Fill-times
tex = GS1.t[-1] + GS2.t[-1] + GS3.t[-1]
tref = GS4.t[-1] + GS5.t[-1] + GS7.t[-1] + readout_time
tend = GS4.t[-1] + GS5.t[-1]
TE_train = tex + n_echo * tref + tend
TR_fill = (TR - n_slices * TE_train) / n_slices  # Round to gradient raster

TR_fill = system.grad_raster_time * round(TR_fill / system.grad_raster_time)
if TR_fill < 0:
    TR_fill = 1e-3
    warnings.warn(f'TR too short, adapted to include all slices to: {1000 * n_slices * (TE_train + TR_fill)} ms')
else:
    print(f'TR fill: {1000 * TR_fill} ms')
delay_TR = make_delay(TR_fill)
delay_end = make_delay(5)

# ======
# CONSTRUCT SEQUENCE
# ======
# Define sequence blocks
for k_ex in range(n_ex):
    for s in range(n_slices):
        rfex.freq_offset = GS_ex.amplitude * slice_thickness * (s - (n_slices - 1) / 2)
        rfref.freq_offset = GS_ref.amplitude * slice_thickness * (s - (n_slices - 1) / 2)
        # Align the phase for off-center slices
        rfex.phase_offset = rfex_phase - 2 * math.pi * rfex.freq_offset * calc_rf_center(rfex)[0]
        rfref.phase_offset = rfref_phase - 2 * math.pi * rfref.freq_offset * calc_rf_center(rfref)[0]

        seq.add_block(GS1)
        seq.add_block(GS2, rfex)
        seq.add_block(GS3, GR3)

        for k_ech in range(n_echo):
            if k_ex >= 0:
                phase_area = phase_areas[k_ech]
            else:
                phase_area = 0

            GP_pre = make_trapezoid(channel='y', system=system, area=phase_area, duration=t_sp, rise_time=dG)
            GP_rew = make_trapezoid(channel='y', system=system, area=-phase_area, duration=t_sp, rise_time=dG)

            seq.add_block(GS4, rfref)
            seq.add_block(GS5, GR5, GP_pre)

            if k_ex >= 0:
                seq.add_block(GR6, adc)
            else:
                seq.add_block(GR6)

            seq.add_block(GS7, GR7, GP_rew)

        seq.add_block(GS4)
        seq.add_block(GS5)
        seq.add_block(delay_TR)

seq.add_block(delay_end)

# ======
# VISUALIZATION
# ======
seq.plot()

ktraj_adc, ktraj, t_excitation, t_refocusing, _ = seq.calculate_kspace()
plt.plot(ktraj.T)  # Plot the entire k-space trajectory
plt.figure()
plt.plot(ktraj[0], ktraj[1], 'b')  # 2D plot
plt.axis('equal')  # Enforce aspect ratio for the correct trajectory display
plt.plot(ktraj_adc[0], ktraj_adc[1], 'r.')
plt.show()

seq.write('haste_pypulseq.seq')



import math
from types import SimpleNamespace
from typing import Tuple, Union

import numpy as np

from pypulseq.make_trap_pulse import make_trapezoid
from pypulseq.opts import Opts


def make_block_pulse(flip_angle: float, bandwidth: float = 0, delay: float = 0, duration: float = 0,
                     freq_offset: float = 0, max_grad: float = 0, max_slew: float = 0, phase_offset: float = 0,
                     return_gz: bool = False, system: Opts = Opts(), slice_thickness: float = 0,
                     time_bw_product: float = 0,
                     use: str = str()) -> Union[SimpleNamespace, Tuple[SimpleNamespace, SimpleNamespace]]:
    """
    Creates a radio-frequency block pulse event and optionally an accompanying slice select trapezoidal gradient event.

    Parameters
    ----------
    flip_angle : float
        Flip angle in radians.
    bandwidth : float, default=0
        Bandwidth in Hertz (hz).
    delay : float, default=0
        Delay in milliseconds (ms) of accompanying slice select trapezoidal event.
    duration : float, default=0
        Duration in milliseconds (ms).
    freq_offset : float, default=0
        Frequency offset in Hertz (Hz).
    max_grad : float, default=0
        Maximum gradient strength of accompanying slice select trapezoidal event.
    max_slew : float, default=0
        Maximum slew rate of accompanying slice select trapezoidal event.
    phase_offset : float, default=0
        Phase offset Hertz (Hz).
    return_gz : bool, default=False
        Boolean flag to indicate if slice-selective gradient has to be returned.
    slice_thickness : float, default=0
        Slice thickness of accompanying slice select trapezoidal event. The slice thickness determines the area of the
        slice select event.
    system : Opts, default=Opts()
        System limits.
    time_bw_product : float, default=0
        Time-bandwidth product.
    use : str, default=str()
        Use of radio-frequency block pulse event. Must be one of 'excitation', 'refocusing' or 'inversion'.

    Returns
    -------
    rf : SimpleNamespace
        Radio-frequency block pulse event.
    gz : SimpleNamespace
        Slice select trapezoidal gradient event accompanying the radio-frequency block pulse event.

    Raises
    ------
    ValueError
        If invalid `use` parameter is passed. Must be one of 'excitation', 'refocusing' or 'inversion'.
        If neither `bandwidth` nor `duration` are passed.
        If `return_gz=True`, and `slice_thickness` is not passed.
    """
    valid_use_pulses = ['excitation', 'refocusing', 'inversion']
    if use is not None and use not in valid_use_pulses:
        raise ValueError(
            f"Invalid use parameter. Must be one of 'excitation', 'refocusing' or 'inversion'. Passed: {use}")

    if duration == 0:
        if time_bw_product > 0:
            duration = time_bw_product / bandwidth
        elif bandwidth > 0:
            duration = 1 / (4 * bandwidth)
        else:
            raise ValueError('Either bandwidth or duration must be defined')

    BW = 1 / (4 * duration)
    N = round(duration / 1e-6)
    t = np.arange(1, N + 1) * system.rf_raster_time
    signal = flip_angle / (2 * np.pi) / duration * np.ones(len(t))

    rf = SimpleNamespace()
    rf.type = 'rf'
    rf.signal = signal
    rf.t = t
    rf.freq_offset = freq_offset
    rf.phase_offset = phase_offset
    rf.dead_time = system.rf_dead_time
    rf.ringdown_time = system.rf_ringdown_time
    rf.delay = delay

    if use != '':
        rf.use = use

    if rf.dead_time > rf.delay:
        rf.delay = rf.dead_time

    if return_gz:
        if slice_thickness < 0:
            raise ValueError('Slice thickness must be provided')

        if max_grad > 0:
            system.max_grad = max_grad
        if max_slew > 0:
            system.max_slew = max_slew

        amplitude = BW / slice_thickness
        area = amplitude * duration
        gz = make_trapezoid(channel='z', system=system, flat_time=duration, flat_area=area)

        if rf.delay > gz.rise_time:
            gz.delay = math.ceil((rf.delay - gz.rise_time) / system.grad_raster_time) * system.grad_raster_time

        if rf.delay < (gz.rise_time + gz.delay):
            rf.delay = gz.rise_time + gz.delay

    if rf.ringdown_time > 0:
        t_fill = np.arange(1, round(rf.ringdown_time / 1e-6) + 1) * 1e-6
        rf.t = np.concatenate((rf.t, (rf.t[-1] + t_fill)))
        rf.signal = np.concatenate((rf.signal, np.zeros(len(t_fill))))

    return rf, gz if return_gz else rf



from types import SimpleNamespace

import numpy as np

from pypulseq.block_to_events import block_to_events
from pypulseq.calc_duration import calc_duration
from pypulseq.compress_shape import compress_shape
from pypulseq.decompress_shape import decompress_shape
from pypulseq.supported_labels import get_supported_labels


def add_block(self, block_index: int, *args: SimpleNamespace) -> None:
    """
    Inserts PyPulseq block of sequence events into `self.dict_block_events` at position `block_index`. Also performs
    gradient checks.

    Parameters
    ----------
    block_index : int
        Index at which `SimpleNamespace` objects have to be inserted into `self.dict_block_events`.
    args : iterable[SimpleNamespace]
        Iterable of `SimpleNamespace` objects to be added to `self.dict_block_events`.

    Raises
    ------
    ValueError
        If trigger event that is passed is of unsupported control event type.
        If delay is set for a gradient even that starts with a non-zero amplitude.
    RuntimeError
        If two consecutive gradients to not have the same amplitude at the connection point.
        If the first gradient in the block does not start with 0.
        If a gradient that doesn't end at zero is not aligned to the block boundary.
    """
    events = block_to_events(args)
    block_duration = calc_duration(*events)
    self.dict_block_events[block_index] = np.zeros(7, dtype=np.int)
    duration = 0

    check_g = {}  # Key-value mapping of index and  pairs of gradients/times
    extensions = []

    for event in events:
        if event.type == 'rf':
            mag = np.abs(event.signal)
            amplitude = np.max(mag)
            mag = np.divide(mag, amplitude)
            # Following line of code is a workaround for numpy's divide functions returning NaN when mathematical
            # edge cases are encountered (eg. divide by 0)
            mag[np.isnan(mag)] = 0
            phase = np.angle(event.signal)
            phase[phase < 0] += 2 * np.pi
            phase /= 2 * np.pi

            mag_shape = compress_shape(mag)
            data = np.insert(mag_shape.data, 0, mag_shape.num_samples)
            mag_id, found = self.shape_library.find(data)
            if not found:
                self.shape_library.insert(mag_id, data)

            phase_shape = compress_shape(phase)
            data = np.insert(phase_shape.data, 0, phase_shape.num_samples)
            phase_id, found = self.shape_library.find(data)
            if not found:
                self.shape_library.insert(phase_id, data)

            use = 0
            use_cases = {'excitation': 1, 'refocusing': 2, 'inversion': 3}
            if hasattr(event, 'use'):
                use = use_cases[event.use]

            data = [amplitude, mag_id, phase_id, event.delay, event.freq_offset, event.phase_offset, event.dead_time,
                    event.ringdown_time, use]
            data_id, found = self.rf_library.find(data)
            if not found:
                self.rf_library.insert(data_id, data)

            self.dict_block_events[block_index][1] = data_id
            duration = max(duration, len(mag) * self.rf_raster_time + event.delay)
        elif event.type == 'grad':
            channel_num = ['x', 'y', 'z'].index(event.channel)
            idx = 2 + channel_num

            check_g[channel_num] = SimpleNamespace()
            check_g[channel_num].idx = idx
            check_g[channel_num].start = np.array((event.delay + min(event.t), event.first))
            check_g[channel_num].stop = np.array(
                (event.delay + max(event.t) + self.system.grad_raster_time, event.last))

            amplitude = max(abs(event.waveform))
            if amplitude > 0:
                g = event.waveform / amplitude
            else:
                g = event.waveform
            shape = compress_shape(g)
            data = np.insert(shape.data, 0, shape.num_samples)
            shape_id, found = self.shape_library.find(data)
            if not found:
                self.shape_library.insert(shape_id, data)
            data = [amplitude, shape_id, event.delay, event.first, event.last]
            grad_id, found = self.grad_library.find(data)
            if not found:
                self.grad_library.insert(grad_id, data, 'g')
            self.dict_block_events[block_index][idx] = grad_id
            duration = max(duration, event.delay + len(g) * self.grad_raster_time)
        elif event.type == 'trap':
            channel_num = ['x', 'y', 'z'].index(event.channel)
            idx = 2 + channel_num

            check_g[channel_num] = SimpleNamespace()
            check_g[channel_num].idx = idx
            check_g[channel_num].start = np.array((0, 0))
            check_g[channel_num].stop = np.array((event.delay + event.rise_time + event.fall_time + event.flat_time, 0))

            data = [event.amplitude, event.rise_time, event.flat_time, event.fall_time, event.delay]
            trap_id, found = self.grad_library.find(data)
            if not found:
                self.grad_library.insert(trap_id, data, 't')
            self.dict_block_events[block_index][idx] = trap_id
            duration = max(duration, event.delay + event.rise_time + event.flat_time + event.fall_time)
        elif event.type == 'adc':
            data = [event.num_samples, event.dwell, max(event.delay, event.dead_time), event.freq_offset,
                    event.phase_offset, event.dead_time]
            adc_id, found = self.adc_library.find(data)
            if not found:
                self.adc_library.insert(adc_id, data)
            self.dict_block_events[block_index][5] = adc_id
            duration = max(duration, event.delay + event.num_samples * event.dwell + event.dead_time)
        elif event.type == 'delay':
            data = [event.delay]
            delay_id, found = self.delay_library.find(data)
            if not found:
                self.delay_library.insert(delay_id, data)
            self.dict_block_events[block_index][0] = delay_id
            duration = max(duration, event.delay)
        elif event.type == 'output' or event.type == 'trigger':
            event_type = ['output', 'trigger'].index(event.type) + 1
            if event_type == 1:
                # Trigger codes supported by the Siemens interpreter as of May 2019
                event_channel = ['osc0', 'osc1', 'ext1'].index(event.channel) + 1
            elif event_type == 2:
                # Trigger codes supported by the Siemens interpreter as of June 2019
                event_channel = ['physio1', 'physio2'].index(event.channel) + 1
            else:
                raise ValueError('Unsupported control event type.')

            data = [event_type, event_channel, event.delay, event.duration]
            trigger_id, found = self.trigger_library.find(data)
            if not found:
                self.trigger_library.insert(trigger_id, data)

            # Now we collect the list of extension objects and we will add it to the event table later
            ext = {'type': self.get_extension_type_ID('TRIGGERS'), 'ref': trigger_id}
            extensions.append(ext)
            duration = max(duration, event.delay + event.duration)
        elif event.type == 'labelset':
            label_id = get_supported_labels().index(event.label) + 1
            data = [event.value, label_id]
            label_id2, found = self.label_set_library.find(data)
            if not found:
                self.label_set_library.insert(label_id2, data)

            ext = {'type': self.get_extension_type_ID('LABELSET'), 'ref': label_id2}
            extensions.append(ext)
        elif event.type == 'labelinc':
            label_id = get_supported_labels().index(event.label) + 1
            data = [event.value, label_id]
            label_id2, found = self.label_inc_library.find(data)
            if not found:
                self.label_inc_library.insert(label_id2, data)

            ext = {'type': self.get_extension_type_ID('LABELINC'), 'ref': label_id2}
            extensions.append(ext)

    # =========
    # ADD EXTENSIONS
    # =========
    if len(extensions) > 0:
        """
        Add extensions now... but it's tricky actually we need to check whether the exactly the same list of extensions 
        already exists, otherwise we have to create a new one... ooops, we have a potential problem with the key 
        mapping then... The trick is that we rely on the sorting of the extension IDs and then we can always find the 
        last one in the list by setting the reference to the next to 0 and then proceed with the other elements.
        """
        sort_idx = np.argsort([e['ref'] for e in extensions])
        extensions = np.take(extensions, sort_idx)
        all_found = True
        extension_id = 0
        for i in range(len(extensions)):
            data = [extensions[i]['type'], extensions[i]['ref'], extension_id]
            extension_id, found = self.extensions_library.find(data)
            all_found = all_found and found
            if not found:
                break

        if not all_found:
            # Add the list
            extension_id = 0
            for i in range(len(extensions)):
                data = [extensions[i]['type'], extensions[i]['ref'], extension_id]
                extension_id, found = self.extensions_library.find(data)
                if not found:
                    self.extensions_library.insert(extension_id, data)

        # Now we add the ID
        self.dict_block_events[block_index][6] = extension_id

    # =========
    # PERFORM GRADIENT CHECKS
    # =========
    for grad_to_check in check_g.values():

        if abs(grad_to_check.start[1]) > self.system.max_slew * self.system.grad_raster_time:
            if grad_to_check.start[0] != 0:
                raise ValueError('No delay allowed for gradients which start with a non-zero amplitude')

            if block_index > 1:
                prev_id = self.dict_block_events[block_index - 1][grad_to_check.idx]
                if prev_id != 0:
                    prev_lib = self.grad_library.get(prev_id)
                    prev_dat = prev_lib['data']
                    prev_type = prev_lib['type']
                    if prev_type == 't':
                        raise RuntimeError(
                            'Two consecutive gradients need to have the same amplitude at the connection point')
                    elif prev_type == 'g':
                        last = prev_dat[4]
                        if abs(last - grad_to_check.start[1]) > self.system.max_slew * self.system.grad_raster_time:
                            raise RuntimeError(
                                'Two consecutive gradients need to have the same amplitude at the connection point')
            else:
                raise RuntimeError('First gradient in the the first block has to start at 0.')

        if grad_to_check.stop[1] > self.system.max_slew * self.system.grad_raster_time and abs(
                grad_to_check.stop[0] - block_duration) > 1e-7:
            raise RuntimeError("A gradient that doesn't end at zero needs to be aligned to the block boundary.")

    eps = np.finfo(np.float).eps
    assert abs(duration - block_duration) < eps
    self.arr_block_durations.append(block_duration)


def get_block(self, block_index: int) -> SimpleNamespace:
    """
    Returns PyPulseq block at `block_index` position in `self.dict_block_events`.

    Parameters
    ----------
    block_index : int
        Index of PyPulseq block to be retrieved from `self.dict_block_events`.

    Returns
    -------
    block : SimpleNamespace
        PyPulseq block at 'block_index' position in `self.dict_block_events`.

    Raises
    ------
    ValueError
        If a trigger event of an unsupported control type is encountered.
        If a label object of an unknown extension ID is encountered.
    """

    block = SimpleNamespace()
    event_ind = self.dict_block_events[block_index]

    if event_ind[0] > 0:  # Delay
        delay = SimpleNamespace()
        delay.type = 'delay'
        delay.delay = self.delay_library.data[event_ind[0]][0]
        block.delay = delay

    if event_ind[1] > 0:  # RF
        block.rf = self.rf_from_lib_data(self.rf_library.data[event_ind[1]])

    # Gradients
    grad_channels = ['gx', 'gy', 'gz']
    for i in range(1, len(grad_channels) + 1):
        if event_ind[2 + (i - 1)] > 0:
            grad, compressed = SimpleNamespace(), SimpleNamespace()
            grad_type = self.grad_library.type[event_ind[2 + (i - 1)]]
            lib_data = self.grad_library.data[event_ind[2 + (i - 1)]]
            grad.type = 'trap' if grad_type == 't' else 'grad'
            grad.channel = grad_channels[i - 1][1]
            if grad.type == 'grad':
                amplitude = lib_data[0]
                shape_id = lib_data[1]
                delay = lib_data[2]
                shape_data = self.shape_library.data[shape_id]
                compressed.num_samples = shape_data[0]
                compressed.data = shape_data[1:]
                g = decompress_shape(compressed)
                grad.waveform = amplitude * g
                grad.t = np.arange(g.size) * self.grad_raster_time
                grad.delay = delay
                if len(lib_data) > 4:
                    grad.first = lib_data[3]
                    grad.last = lib_data[4]
                else:
                    grad.first = grad.waveform[0]
                    grad.last = grad.waveform[-1]
            else:
                if max(lib_data.shape) < 5:  # added by GT
                    grad.amplitude, grad.rise_time, grad.flat_time, grad.fall_time = [lib_data[x] for x in range(4)]
                    grad.delay = 0
                else:
                    grad.amplitude, grad.rise_time, grad.flat_time, grad.fall_time, grad.delay = [lib_data[x] for x in
                                                                                                  range(5)]
                grad.area = grad.amplitude * (grad.flat_time + grad.rise_time / 2 + grad.fall_time / 2)
                grad.flat_area = grad.amplitude * grad.flat_time
            setattr(block, grad_channels[i - 1], grad)
    # ADC
    if event_ind[5] > 0:
        lib_data = self.adc_library.data[event_ind[5]]
        if len(lib_data) < 6:
            lib_data = np.append(lib_data, 0)

        adc = SimpleNamespace()
        adc.num_samples, adc.dwell, adc.delay, adc.freq_offset, adc.phase_offset, adc.dead_time = [lib_data[x] for x in
                                                                                                   range(6)]
        adc.num_samples = int(adc.num_samples)
        adc.type = 'adc'
        block.adc = adc

    # Triggers
    if event_ind[6] > 0:
        # We have extensions - triggers, labels, etc.
        next_ext_id = event_ind[6]
        while next_ext_id != 0:
            ext_data = self.extensions_library.data[next_ext_id]
            # Format: ext_type, ext_id, next_ext_id
            ext_type = self.get_extension_type_string(ext_data[0])

            if ext_type == 'TRIGGERS':
                trigger_types = ['output', 'trigger']
                data = self.trigger_library.data[ext_data[1]]
                trigger = SimpleNamespace()
                trigger.type = trigger_types[int(data[0])]
                if data[0] == 0:
                    trigger_channels = ['osc0', 'osc1', 'ext1']
                    trigger.channel = trigger_channels[int(data[1])]
                elif data[0] == 1:
                    trigger_channels = ['physio1', 'physio2']
                    trigger.channel = trigger_channels[int(data[1])]
                else:
                    raise ValueError('Unsupported trigger event type')

                trigger.delay = data[2]
                trigger.duration = data[3]
                # Allow for multiple triggers per block
                if hasattr(block, 'trigger'):
                    block.trigger[len(block.trigger)] = trigger
                else:
                    block.trigger = {0: trigger}
            elif ext_type == 'LABELSET' or ext_type == 'LABELINC':
                label = SimpleNamespace()
                label.type = ext_type.lower()
                supported_labels = get_supported_labels()
                if ext_type == 'LABELSET':
                    data = self.label_set_library.data[ext_data[1]]
                else:
                    data = self.label_inc_library.data[ext_data[1]]

                label.label = supported_labels[data[1] - 1]
                label.value = data[0]
                # Allow for multiple labels per block
                if hasattr(block, 'label'):
                    block.label[len(block.label)] = label
                else:
                    block.label = {0: label}
            else:
                raise RuntimeError(f'Unknown extension ID {ext_data[0]}')

            next_ext_id = ext_data[2]

    return block



"""
A very basic UTE-like sequence, without ramp-sampling, ramp-RF. Achieves TE in the range of 300-400 us
"""
from copy import copy

import numpy as np
from matplotlib import pyplot as plt

import pypulseq as pp

# ======
# SETUP
# ======
seq = pp.Sequence()  # Create a new sequence object
# Define FOV and resolution
fov = 250e-3
Nx = 256
alpha = 10  # Flip angle
slice_thickness = 3e-3  # Slice thickness
TR = 10e-3  # TR
Nr = 128  # Number of radial spokes
delta = 2 * np.pi / Nr  # AAngular increment; try golden angle pi*(3-5^0.5) or 0.5 of i
ro_duration = 2.4e-3  # Read-out time: controls RO bandwidth and T2-blurring
ro_os = 2  # Oversampling
ro_asymmetry = 0.97  # 0: Fully symmetric; 1: half-echo
minRF_to_ADC_time = 50e-6  # Defines TE together with the RO asymmetry

rf_spoiling_inc = 117  # RF spoiling increment

# Set system limits
system = pp.Opts(max_grad=28, grad_unit='mT/m', max_slew=100, slew_unit='T/m/s', rf_ringdown_time=20e-6,
                 rf_dead_time=100e-6, adc_dead_time=10e-6)

# ======
# CREATE EVENTS
# ======
# Create alpha-degree slice selection pulse and gradient
rf, gz, gz_reph = pp.make_sinc_pulse(flip_angle=alpha * np.pi / 180, duration=1e-3, slice_thickness=slice_thickness,
                                     apodization=0.5, time_bw_product=2, center_pos=1, system=system, return_gz=True)

# Align RO asymmetry to ADC samples
Nxo = np.round(ro_os * Nx)
ro_asymmetry = np.round(ro_asymmetry * Nxo / 2) / Nxo * 2

# Define other gradients and ADC events
delta_k = 1 / fov / (1 + ro_asymmetry)
ro_area = Nx * delta_k
gx = pp.make_trapezoid(channel='x', flat_area=ro_area, flat_time=ro_duration, system=system)
adc = pp.make_adc(num_samples=Nxo, duration=gx.flat_time, delay=gx.rise_time, system=system)
adc.delay = adc.delay - 0.5 * adc.dwell  # compensate for the 0.5 samples shift
gx_pre = pp.make_trapezoid(channel='x', area=-(gx.area - ro_area) / 2 - ro_area / 2 * (1 - ro_asymmetry), system=system)

# Gradient spoiling
gx_spoil = pp.make_trapezoid(channel='x', area=0.2 * Nx * delta_k, system=system)

# Calculate timing
TE = gz.fall_time + pp.calc_duration(gx_pre, gz_reph) + gx.rise_time + adc.dwell * Nxo / 2 * (1 - ro_asymmetry)
delay_TR = np.ceil((TR - pp.calc_duration(gx_pre, gz_reph) - pp.calc_duration(gz) - pp.calc_duration(
    gx)) / seq.grad_raster_time) * seq.grad_raster_time
assert np.all(delay_TR >= pp.calc_duration(gx_spoil))

print(f'TE = {TE * 1e6:.0f} us')

if pp.calc_duration(gz_reph) > pp.calc_duration(gx_pre):
    gx_pre.delay = pp.calc_duration(gz_reph) - pp.calc_duration(gx_pre)

rf_phase = 0
rf_inc = 0

# ======
# CONSTRUCT SEQUENCE
# ======
for i in range(Nr):
    for c in range(2):
        rf.phase_offset = rf_phase / 180 * np.pi
        adc.phase_offset = rf_phase / 180 * np.pi
        rf_inc = np.mod(rf_inc + rf_spoiling_inc, 360.0)
        rf_phase = np.mod(rf_phase + rf_inc, 360.0)

        gz.amplitude = -gz.amplitude  # Alternate GZ amplitude
        gz_reph.amplitude = -gz_reph.amplitude

        seq.add_block(rf, gz)
        phi = delta * i

        gpc = copy(gx_pre)
        gps = copy(gx_pre)
        gpc.amplitude = gx_pre.amplitude * np.cos(phi)
        gps.amplitude = gx_pre.amplitude * np.sin(phi)
        gps.channel = 'y'

        grc = copy(gx)
        grs = copy(gx)
        grc.amplitude = gx.amplitude * np.cos(phi)
        grs.amplitude = gx.amplitude * np.sin(phi)
        grs.channel = 'y'

        gsc = copy(gx_spoil)
        gss = copy(gx_spoil)
        gsc.amplitude = gx_spoil.amplitude * np.cos(phi)
        gss.amplitude = gx_spoil.amplitude * np.sin(phi)
        gss.channel = 'y'

        seq.add_block(gpc, gps, gz_reph)
        seq.add_block(grc, grs, adc)
        seq.add_block(gsc, gss, pp.make_delay(delay_TR))

# ======
# VISUALIZATION
# ======
seq.plot()

# Plot gradients to check for gaps and optimality of the timing
gw = seq.gradient_waveforms()
plt.plot(gw.T)  # Plot the entire gradient shape

# Trajectory calculation
ktraj_adc, ktraj, t_excitation, t_refocusing, t_adc = seq.calculate_kspace()

# Plot k-spaces
time_axis = np.arange(ktraj.shape[1]) * seq.grad_raster_time
plt.figure()
plt.plot(time_axis, ktraj.T)  # Plot the entire k-space trajectory
plt.plot(t_adc, ktraj_adc[0], '.')  # Sampling points on the kx-axis
plt.figure()
plt.plot(ktraj[0], ktraj[1], 'b')  # 2D plot
plt.axis('equal')  # Enforce aspect ratio for the correct trajectory display
plt.plot(ktraj_adc[0], ktraj_adc[1], 'r.')  # Plot the sampling points

plt.show()

seq.set_definition('FOV', [fov, fov, slice_thickness])
seq.set_definition('Name', 'UTE')

seq.write('ute_pypulseq.seq')



from typing import Tuple


def get_supported_labels() -> Tuple[str, str, str, str, str, str, str, str, str, str, str, str]:
    """
    Returns
    -------
    tuple
        Tuple of supported labels.
    """
    return 'SLC', 'SEG', 'REP', 'AVG', 'SET', 'ECO', 'PHS', 'LIN', 'PAR', 'NAV', 'REV', 'SMS'



import math

import numpy as np
from matplotlib import pyplot as plt

import pypulseq as pp

# ======
# SETUP
# ======
seq = pp.Sequence()  # Create a new sequence object
# Define FOV and resolution
fov = 256e-3
Nx = 64
Ny = 64

# Set system limits
system = pp.Opts(max_grad=32, grad_unit='mT/m', max_slew=130, slew_unit='T/m/s', rf_ringdown_time=30e-6,
                 rf_dead_time=100e-6, adc_dead_time=20e-6)

# ======
# CREATE EVENTS
# ======
# Create 90 degree slice selection pulse and gradient
rf, gz, _ = pp.make_sinc_pulse(flip_angle=np.pi / 2, system=system, duration=3e-3, slice_thickness=3e-3,
                               apodization=0.5, time_bw_product=4, return_gz=True)

# Define other gradients and ADC events
delta_k = 1 / fov
k_width = Nx * delta_k
readout_time = 3.2e-4
gx = pp.make_trapezoid(channel='x', system=system, flat_area=k_width, flat_time=readout_time)
adc = pp.make_adc(num_samples=Nx, system=system, duration=gx.flat_time, delay=gx.rise_time)

# Pre-phasing gradients
pre_time = 8e-4
gz_reph = pp.make_trapezoid(channel='z', system=system, area=-gz.area / 2, duration=pre_time)
# Do not need minus for in-plane prephasers because of the spin-echo (position reflection in k-space)
gx_pre = pp.make_trapezoid(channel='x', system=system, area=gx.area / 2 - delta_k / 2, duration=pre_time)
gy_pre = pp.make_trapezoid(channel='y', system=system, area=Ny / 2 * delta_k, duration=pre_time)

# Phase blip in shortest possible time
dur = math.ceil(2 * math.sqrt(delta_k / system.max_slew) / 10e-6) * 10e-6
gy = pp.make_trapezoid(channel='y', system=system, area=delta_k, duration=dur)

# Refocusing pulse with spoiling gradients
rf180, _ = pp.make_block_pulse(flip_angle=np.pi, system=system, duration=500e-6, use='refocusing')
gz_spoil = pp.make_trapezoid(channel='z', system=system, area=gz.area * 2, duration=3 * pre_time)

# Calculate delay time
TE = 60e-3
duration_to_center = (Nx / 2 + 0.5) * pp.calc_duration(gx) + Ny / 2 * pp.calc_duration(gy)
rf_center_incl_delay = rf.delay + pp.calc_rf_center(rf)[0]
rf180_center_incl_delay = rf180.delay + pp.calc_rf_center(rf180)[0]
delay_TE1 = TE / 2 - pp.calc_duration(gz) + rf_center_incl_delay - pre_time - pp.calc_duration(
    gz_spoil) - rf180_center_incl_delay
delay_TE2 = TE / 2 - pp.calc_duration(rf180) + rf180_center_incl_delay - pp.calc_duration(gz_spoil) - duration_to_center

# ======
# CONSTRUCT SEQUENCE
# ======
# Define sequence blocks
seq.add_block(rf, gz)
seq.add_block(gx_pre, gy_pre, gz_reph)
seq.add_block(pp.make_delay(delay_TE1))
seq.add_block(gz_spoil)
seq.add_block(rf180)
seq.add_block(gz_spoil)
seq.add_block(pp.make_delay(delay_TE2))
for i in range(Ny):
    seq.add_block(gx, adc)  # Read one line of k-space
    seq.add_block(gy)  # Phase blip
    gx.amplitude = -gx.amplitude  # Reverse polarity of read gradient
seq.add_block(pp.make_delay(1e-4))

# ======
# VISUALIZATION
# ======
seq.plot()

# Calculate trajectory
ktraj_adc, ktraj, t_excitation, t_refocusing, t_adc = seq.calculate_kspace()

# Plot k-spaces
time_axis = np.arange(1, ktraj.shape[1] + 1) * system.grad_raster_time
plt.figure()
plt.plot(time_axis, ktraj.T)  # Plot entire k-space trajectory
plt.plot(t_adc, ktraj_adc[0], '.')  # Plot sampling points on kx-axis
plt.figure()
plt.plot(ktraj[0], ktraj[1], 'b', ktraj_adc[0], ktraj_adc[1], 'r.')  # 2D plot
plt.axis('equal')
plt.show()

seq.write('epi_se_pypulseq.seq')

# Sanity checks
TE_check = (t_refocusing[0] - t_excitation[0]) * 2
print(f'Intended TE = {TE * 1e3:.03f} ms, actual spin echo TE = {TE_check * 1e3:.03f} ms')



from types import SimpleNamespace
from typing import Tuple

import numpy as np


def calc_rf_center(rf: SimpleNamespace) -> Tuple[float, float]:
    """
    Calculate the time point of the effective rotation calculated as the peak of the radio-frequency amplitude for the
    shaped pulses and the center of the pulse for the block pulses. Zero padding in the radio-frequency pulse is
    considered as a part of the shape. Delay field of the radio-frequency object is not taken into account.

    Parameters
    ----------
    rf : SimpleNamespace
        Radio-frequency pulse event.

    Returns
    -------
    time_center : float
        Time point of the center of the radio-frequency pulse.
    id_center : float
        Corresponding position of `time_center` in the radio-frequency pulse's envelope.
    """
    eps = np.finfo(float).eps
    for first, x in enumerate(rf.signal):
        if abs(x) > eps:
            break

    for last, x in enumerate(rf.signal[::-1]):
        if abs(x) > eps:
            break

    # Detect the excitation peak: we traverse over in-place reverse of rf.signal; we want index from the ending
    last = len(rf.signal) - last - 1
    rf_min = min(abs(rf.signal[first:last + 1]))
    rf_max = max(abs(rf.signal[first:last + 1]))
    id_center = np.argmax(abs(rf.signal[first:last + 1]))
    if rf_max - rf_min <= eps:
        id_center = round((last - first + 1) / 2) - 1

    time_center = rf.t[first + id_center]

    return time_center, id_center



import math
from types import SimpleNamespace

from pypulseq.opts import Opts


def make_trapezoid(channel: str, amplitude: float = 0, area: float = None, delay: float = 0, duration: float = 0,
                   flat_area: float = 0, flat_time: float = 0, max_grad: float = 0, max_slew: float = 0,
                   rise_time: float = 0, system: Opts = Opts()) -> SimpleNamespace:
    """
    Creates a trapezoidal gradient event.

    Parameters
    ----------
    channel : str, optional
        Orientation of trapezoidal gradient event. Must be one of `x`, `y` or `z`.
    amplitude : float, optional, default=0
        Amplitude.
    area : float, optional, default=None
        Area.
    delay : float, optional, default=0
        Delay in milliseconds (ms).
    duration : float, optional, default=0
        Duration in milliseconds (ms).
    flat_area : float, optional, default=0
        Flat area.
    flat_time : float, optional, default=0
        Flat duration in milliseconds (ms).
    max_grad : float, optional, default=0
        Maximum gradient strength.
    max_slew : float, optional, default=0
        Maximum slew rate.
    rise_time : float, optional, default=0
        Rise time in milliseconds (ms).
    system : Opts, optional, default=Opts()
        System limits.

    Returns
    -------
    grad : SimpleNamespace
        Trapezoidal gradient event created based on the supplied parameters.

    Raises
    ------
    ValueError
        - If none of `area`, `flat_area` and `amplitude` are passed
        - If requested area is too large for this gradient
        - Amplitude violation
    """
    if channel not in ['x', 'y', 'z']:
        raise ValueError(f"Invalid channel. Must be one of `x`, `y` or `z`. Passed: {channel}")

    if max_grad <= 0:
        max_grad = system.max_grad

    if max_slew <= 0:
        max_slew = system.max_slew

    if rise_time <= 0:
        rise_time = system.rise_time

    if area is None and flat_area == 0 and amplitude == 0:
        raise ValueError("Must supply either 'area', 'flat_area' or 'amplitude'.")

    if flat_time != 0:
        if amplitude != 0:
            amplitude2 = amplitude
        else:
            amplitude2 = flat_area / flat_time

        if rise_time == 0:
            rise_time = abs(amplitude2) / max_slew
            rise_time = math.ceil(rise_time / system.grad_raster_time) * system.grad_raster_time
        fall_time, flat_time = rise_time, flat_time
    elif duration > 0:
        amplitude2 = amplitude
        if amplitude == 0:
            if rise_time == 0:
                dC = 1 / abs(2 * max_slew) + 1 / abs(2 * max_slew)
                possible = duration ** 2 > 4 * abs(area) * dC
                amplitude2 = (duration - math.sqrt(duration ** 2 - 4 * abs(area) * dC)) / (2 * dC)
            else:
                amplitude2 = area / (duration - rise_time)
                possible = duration > 2 * rise_time and abs(amplitude2) < max_grad

            if not possible:
                raise ValueError('Requested area is too large for this gradient')

        if rise_time == 0:
            rise_time = math.ceil(
                abs(amplitude2) / max_slew / system.grad_raster_time) * system.grad_raster_time
            if rise_time == 0:
                rise_time = system.grad_raster_time

        fall_time = rise_time
        flat_time = duration - rise_time - fall_time

        if amplitude == 0:
            amplitude2 = area / (rise_time / 2 + fall_time / 2 + flat_time)
    else:
        if area is None:
            raise ValueError('Must supply a duration')
        else:
            rise_time = math.ceil(math.sqrt(abs(area) / max_slew) / system.grad_raster_time) * system.grad_raster_time
            amplitude2 = area / rise_time
            t_eff = rise_time

            if abs(amplitude2) > max_grad:
                t_eff = math.ceil(abs(area) / max_grad / system.grad_raster_time) * system.grad_raster_time
                amplitude2 = area / t_eff
                rise_time = math.ceil(
                    abs(amplitude2) / max_slew / system.grad_raster_time) * system.grad_raster_time

            flat_time = t_eff - rise_time
            fall_time = rise_time

    if abs(amplitude2) > max_grad:
        raise ValueError("Amplitude violation")

    grad = SimpleNamespace()
    grad.type = 'trap'
    grad.channel = channel
    grad.amplitude = amplitude2
    grad.rise_time = rise_time
    grad.flat_time = flat_time
    grad.fall_time = fall_time
    grad.area = amplitude2 * (flat_time + rise_time / 2 + fall_time / 2)
    grad.flat_area = amplitude2 * flat_time
    grad.delay = delay
    grad.first = 0
    grad.last = 0

    return grad



# inserted for trigger support by mveldmann

from types import SimpleNamespace

from pypulseq.opts import Opts


def make_trigger(channel: str, delay: float = 0, duration: float = 0, system: Opts = Opts()) -> SimpleNamespace:
    """
    Creates a trigger event.

    Parameters
    ----------
    channel : str
        Must be one of 'physio1' or 'physio2'.
    delay : float, default=0
        Delay in seconds
    duration: float, default=0
        Duration in seconds.
    system : Opts, default=Opts()
        System limits.

    Returns
    -------
    trigger : SimpleNamespace
        Trigger event.

    Raises
    ------
    ValueError
        If invalid `channel` is passed. Must be one of 'physio1' or 'physio2'.
    """

    if channel not in ['physio1', 'physio2']:
        raise ValueError(f"Channel {channel} is invalid. Must be one of 'physio1' or 'physio2'.")

    trigger = SimpleNamespace()
    trigger.type = 'trigger'
    trigger.channel = channel
    trigger.delay = delay
    trigger.duration = duration
    if trigger.duration <= system.grad_raster_time:
        trigger.duration = system.grad_raster_time

    return trigger



import math

import numpy as np
from matplotlib import pyplot as plt

import pypulseq as pp

# ======
# SETUP
# ======
# Create a new sequence object
seq = pp.Sequence()

# Define FOV and resolution
fov = 256e-3
Nx = 256
Ny = 256
alpha = 10  # flip angle
slice_thickness = 3e-3  # slice
# TE = np.array([7.38]) * 1e-3  # give a vector here to have multiple TEs (e.g. for field mapping)
TE = np.array([4.3e-3])
TR = 10e-3

rf_spoiling_inc = 117  # RF spoiling increment

system = pp.Opts(max_grad=28, grad_unit='mT/m', max_slew=150, slew_unit='T/m/s', rf_ringdown_time=20e-6,
                 rf_dead_time=100e-6, adc_dead_time=10e-6)

# ======
# CREATE EVENTS
# ======
rf, gz, gzr = pp.make_sinc_pulse(flip_angle=alpha * math.pi / 180, duration=3e-3, slice_thickness=slice_thickness,
                                 apodization=0.5, time_bw_product=4, system=system, return_gz=True)
# Define other gradients and ADC events
delta_k = 1 / fov
gx = pp.make_trapezoid(channel='x', flat_area=Nx * delta_k, flat_time=3.2e-3, system=system)
adc = pp.make_adc(num_samples=Nx, duration=gx.flat_time, delay=gx.rise_time, system=system)
gx_pre = pp.make_trapezoid(channel='x', area=-gx.area / 2, duration=1e-3, system=system)
gz_reph = pp.make_trapezoid(channel='z', area=-gz.area / 2, duration=1e-3, system=system)
phase_areas = (np.arange(Ny) - Ny / 2) * delta_k

# gradient spoiling
gx_spoil = pp.make_trapezoid(channel='x', area=2 * Nx * delta_k, system=system)
gz_spoil = pp.make_trapezoid(channel='z', area=4 / slice_thickness, system=system)

# Calculate timing
delay_TE = np.ceil((TE - pp.calc_duration(gx_pre) - gz.fall_time - gz.flat_time / 2 - pp.calc_duration(
    gx) / 2) / seq.grad_raster_time) * seq.grad_raster_time
delay_TR = np.ceil((TR - pp.calc_duration(gz) - pp.calc_duration(gx_pre) - pp.calc_duration(
    gx) - delay_TE) / seq.grad_raster_time) * seq.grad_raster_time

assert np.all(delay_TE >= 0)
assert np.all(delay_TR >= pp.calc_duration(gx_spoil, gz_spoil))

rf_phase = 0
rf_inc = 0

# ======
# CONSTRUCT SEQUENCE
# ======
# Loop over phase encodes and define sequence blocks
for i in range(Ny):
    for j in range(len(TE)):
        rf.phase_offset = rf_phase / 180 * np.pi
        adc.phase_offset = rf_phase / 180 * np.pi
        rf_inc = divmod(rf_inc + rf_spoiling_inc, 360.0)[1]
        rf_phase = divmod(rf_phase + rf_inc, 360.0)[1]

        seq.add_block(rf, gz)
        gy_pre = pp.make_trapezoid(channel='y', area=phase_areas[i], duration=pp.calc_duration(gx_pre), system=system)
        seq.add_block(gx_pre, gy_pre, gz_reph)
        seq.add_block(pp.make_delay(delay_TE[j]))
        seq.add_block(gx, adc)
        gy_pre.amplitude = -gy_pre.amplitude
        seq.add_block(pp.make_delay(delay_TR[j]), gx_spoil, gy_pre, gz_spoil)

ok, error_report = seq.check_timing()  # Check whether the timing of the sequence is correct
if ok:
    print('Timing check passed successfully')
else:
    print('Timing check failed. Error listing follows:')
    [print(e) for e in error_report]

# ======
# VISUALIZATION
# ======
seq.plot()

# Trajectory calculation and plotting
ktraj_adc, ktraj, t_excitation, t_refocusing, t_adc = seq.calculate_kspace()
time_axis = np.arange(1, ktraj.shape[1] + 1) * system.grad_raster_time
plt.figure()
plt.plot(time_axis, ktraj.T)  # Plot the entire k-space trajectory
plt.plot(t_adc, ktraj_adc[0], '.')  # Plot sampling points on the kx-axis
plt.figure()
plt.plot(ktraj[0], ktraj[1], 'b')  # 2D plot
plt.axis('equal')  # Enforce aspect ratio for the correct trajectory display
plt.plot(ktraj_adc[0], ktraj_adc[1], 'r.')  # Plot  sampling points
plt.show()

# Prepare the sequence output for the scanner
seq.set_definition('FOV', [fov, fov, slice_thickness])
seq.set_definition('Name', 'gre')

seq.write('gre_pypulseq.seq')

# Very optional slow step, but useful for testing during development e.g. for the real TE, TR or for staying within
# slew-rate limits
rep = seq.test_report()
print(rep)



from pathlib import Path

path_version = Path(__file__).parent.parent / 'VERSION'
with open(str(path_version), 'r') as version_file:
    version = version_file.read().strip().split('.')
    major, minor, revision = [int(v) for v in version]

from pypulseq.SAR.SAR_calc import calc_SAR
from pypulseq.Sequence.sequence import Sequence
from pypulseq.add_gradients import add_gradients
from pypulseq.align import align
from pypulseq.calc_duration import calc_duration
from pypulseq.calc_ramp import calc_ramp
from pypulseq.calc_rf_center import calc_rf_center
from pypulseq.make_adc import make_adc
from pypulseq.make_arbitrary_rf import make_arbitrary_rf
from pypulseq.make_block_pulse import make_block_pulse
from pypulseq.make_delay import make_delay
from pypulseq.make_digital_output_pulse import make_digital_output_pulse
from pypulseq.make_extended_trapezoid import make_extended_trapezoid
from pypulseq.make_extended_trapezoid_area import make_extended_trapezoid_area
from pypulseq.make_gauss_pulse import make_gauss_pulse
from pypulseq.make_label import make_label
from pypulseq.make_sinc_pulse import make_sinc_pulse
from pypulseq.make_trap_pulse import make_trapezoid
from pypulseq.make_trigger import make_trigger
from pypulseq.opts import Opts
from pypulseq.points_to_waveform import points_to_waveform
from pypulseq.split_gradient import split_gradient
from pypulseq.split_gradient_at import split_gradient_at
from pypulseq.supported_labels import get_supported_labels
from pypulseq.traj_to_grad import traj_to_grad



from types import SimpleNamespace
from typing import Tuple


def block_to_events(args: Tuple[SimpleNamespace, ...]) -> Tuple[SimpleNamespace, ...]:
    """
    Converts `args` from a block to a list of events. If `args` is already a list of event(s), returns it unmodified.

    Parameters
    ----------
    args : list[SimpleNamespace]
        Block to be converted into a list of events, or list of events.

    Returns
    -------
    events : list[SimpleNamespace]
        List of events comprising `args` if it was a block, otherwise `args` unmodified.
    """
    if len(args) == 1:  # args is a tuple consisting either a block or a single event
        x = args[0]
        attrs = vars(x).keys()
        children = [getattr(x, a) for a in attrs]
        if all([isinstance(c, (SimpleNamespace, dict)) for c in children]):  # args is a block of events
            events = list(x.__dict__.values())

            # Are any of the events labels? If yes, extract them from dict()
            for e in events:
                if isinstance(e, dict):
                    events.remove(e)
                    events.extend(e.values())
        else:  # args is a single event
            events = [x]
    else:  # args is a tuple of events
        events = args

    return events



from types import SimpleNamespace
from typing import Union

from pypulseq import supported_labels


def make_label(type: str, label: str, value: Union[bool, float, int]) -> SimpleNamespace:
    """
    Parameters
    ----------
    type : str
        Label type. Must be one of 'SET' or 'INC'.
    label : str
        Must be one of 'SLC', 'SEG', 'REP', 'AVG', 'SET', 'ECO', 'PHS', 'LIN', 'PAR', 'NAV', 'REV', or 'SMS'.
    value : bool, float or int
        Label value.

    Returns
    -------
    out : SimpleNamespace
        Label object.

    Raises
    ------
    ValueError
        If a valid `label` was not passed. Must be one of 'SLC', 'SEG', 'REP', 'AVG', 'SET', 'ECO', 'PHS', 'LIN', 'PAR',
                                                                                                NAV', 'REV', or 'SMS'.
        If a valid `type` was not passed. Must be one of 'SET' or 'INC'.
        If `value` was not a valid numerical or logical value.
    """
    arr_supported_labels = supported_labels.get_supported_labels()

    if label not in arr_supported_labels:
        raise ValueError("Invalid label. Must be one of 'SLC', 'SEG', 'REP', 'AVG', 'SET', 'ECO', 'PHS', 'LIN', 'PAR', "
                         "NAV', 'REV', or 'SMS'.")
    if type not in ['SET', 'INC']:
        raise ValueError("Invalid type. Must be one of 'SET' or 'INC'.")
    if not isinstance(value, (bool, float, int)):
        raise ValueError('Must supply a valid numerical or logical value.')

    out = SimpleNamespace()
    if type == 'SET':
        out.type = 'labelset'
    elif type == 'INC':
        out.type = 'labelinc'

    out.label = label
    out.value = value

    return out



from types import SimpleNamespace

from pypulseq.opts import Opts


def make_digital_output_pulse(channel: str, delay: float = 0, duration: float = 0,
                              system: Opts = Opts()) -> SimpleNamespace:
    """
    Create a digital output pulse event a.k.a. trigger. Creates an output trigger event on a given channel with optional
    given delay and duration.

    Parameters
    ----------
    channel : str
        Must be one of 'osc0','osc1', or 'ext1'.
    delay : float, optional, default=0
        Delay, in millis.
    duration : float, optional, default=0
        Duration of trigger event, in millis.
    system : Opts, optional, default=Opts()
        System limits.

    Returns
    ------
    trig : SimpleNamespace
        Trigger event.

    Raises
    ------
    ValueError
        If `channel` is invalid. Must be one of 'osc0','osc1', or 'ext1'.
    """
    if channel not in ['osc0', 'osc1', 'ext1']:
        raise ValueError(f"Channel {channel} is invalid. Must be one of 'osc0','osc1', or 'ext1'.")

    trig = SimpleNamespace()
    trig.type = 'output'
    trig.channel = channel
    trig.delay = delay
    trig.duration = duration
    if trig.duration <= system.grad_raster_time:
        trig.duration = system.grad_raster_time

    return trig



"""
This is an experimental high-performance EPI sequence which uses split gradients to overlap blips with the readout 
gradients combined with ramp-sampling.
"""
import math

import numpy as np
from matplotlib import pyplot as plt

import pypulseq as pp

# ======
# SETUP
# ======
seq = pp.Sequence()  # Create a new sequence object
# Define FOV and resolution
fov = 250e-3
Nx = 64
Ny = 64
slice_thickness = 3e-3  # Clice thickness
n_slices = 3
TE = 40e-3

pe_enable = 1  # Flag to quickly disable phase encoding (1/0) as needed for the delay calibration
ro_os = 1  # Oversampling factor
readout_time = 4.2e-4  # Readout bandwidth
part_fourier_factor = 0.75  # Partial Fourier factor: 1: full sampling; 0: start with ky=0

t_RF_ex = 2e-3
t_RF_ref = 2e-3
spoil_factor = 1.5  # Spoiling gradient around the pi-pulse (rf180)

# Set system limits
system = pp.Opts(max_grad=32, grad_unit='mT/m', max_slew=130, slew_unit='T/m/s', rf_ringdown_time=30e-6,
                 rf_dead_time=100e-6)

# ======
# CREATE EVENTS
# ======
# Create fat-sat pulse
B0 = 2.89
sat_ppm = -3.45
sat_freq = sat_ppm * 1e-6 * B0 * system.gamma
rf_fs = pp.make_gauss_pulse(flip_angle=110 * np.pi / 180, system=system, duration=8e-3, bandwidth=abs(sat_freq),
                            freq_offset=sat_freq)
gz_fs = pp.make_trapezoid(channel='z', system=system, delay=pp.calc_duration(rf_fs), area=1 / 1e-4)

# Create 90 degree slice selection pulse and gradient
rf, gz, gz_reph = pp.make_sinc_pulse(flip_angle=np.pi / 2, system=system, duration=t_RF_ex,
                                     slice_thickness=slice_thickness, apodization=0.5, time_bw_product=4,
                                     return_gz=True)

# Create 90 degree slice refocusing pulse and gradients
rf180, gz180, _ = pp.make_sinc_pulse(flip_angle=np.pi, system=system, duration=t_RF_ref,
                                     slice_thickness=slice_thickness, apodization=0.5, time_bw_product=4,
                                     phase_offset=np.pi / 2, use='refocusing', return_gz=True)
_, gzr1_t, gzr1_a = pp.make_extended_trapezoid_area(channel='z', Gs=0, Ge=gz180.amplitude, A=spoil_factor * gz.area,
                                                    system=system)
_, gzr2_t, gzr2_a = pp.make_extended_trapezoid_area(channel='z', Gs=gz180.amplitude, Ge=0,
                                                    A=-gz_reph.area + spoil_factor * gz.area, system=system)
if gz180.delay > (gzr1_t[3] - gz180.rise_time):
    gz180.delay -= gzr1_t[3] - gz180.rise_time
else:
    rf180.delay += (gzr1_t[3] - gz180.rise_time) - gz180.delay
gz180n = pp.make_extended_trapezoid(channel='z', system=system,
                                    times=np.array([*gzr1_t, *gzr1_t[3] + gz180.flat_time + gzr2_t]) + gz180.delay,
                                    amplitudes=np.array([*gzr1_a, *gzr2_a]))

# Define the output trigger to play out with every slice excitation
trig = pp.make_digital_output_pulse(channel='osc0', duration=100e-6)

# Define other gradients and ADC events
delta_k = 1 / fov
k_width = Nx * delta_k

# Phase blip in shortest possible time
# Round up the duration to 2x gradient raster time
blip_duration = np.ceil(2 * np.sqrt(delta_k / system.max_slew) / 10e-6 / 2) * 10e-6 * 2
# Use negative blips to save one k-space line on our way to center of k-space
gy = pp.make_trapezoid(channel='y', system=system, area=-delta_k, duration=blip_duration)

# Readout gradient is a truncated trapezoid with dead times at the beginning and at the end each equal to a half of blip
# duration. The area between the blips should be defined by k_width. We do a two-step calculation: we first increase the
# area assuming maximum slew rate and then scale down the amplitude to fix the area
extra_area = blip_duration / 2 * blip_duration / 2 * system.max_slew
gx = pp.make_trapezoid(channel='x', system=system, area=k_width + extra_area, duration=readout_time + blip_duration)
actual_area = gx.area - gx.amplitude / gx.rise_time * blip_duration / 2 * blip_duration / 2 / 2
actual_area -= gx.amplitude / gx.fall_time * blip_duration / 2 * blip_duration / 2 / 2
gx.amplitude = gx.amplitude / actual_area * k_width
gx.area = gx.amplitude * (gx.flat_time + gx.rise_time / 2 + gx.fall_time / 2)
gx.flat_area = gx.amplitude * gx.flat_time

# Calculate ADC
# We use ramp sampling, so we have to calculate the dwell time and the number of samples, which will be quite different
# from Nx and readout_time/Nx, respectively.
adc_dwell_nyquist = delta_k / gx.amplitude / ro_os
# Round-down dwell time to 100 ns
adc_dwell = math.floor(adc_dwell_nyquist * 1e7) * 1e-7
adc_samples = math.floor(readout_time / adc_dwell / 4) * 4  # Number of samples on Siemens needs to be divisible by 4
adc = pp.make_adc(num_samples=adc_samples, dwell=adc_dwell, delay=blip_duration / 2)
# Realign the ADC with respect to the gradient
time_to_center = adc_dwell * ((adc_samples - 1) / 2 + 0.5)  # Supposedly Siemens samples at center of dwell period
# Adjust delay to align the trajectory with the gradient. We have to align the delay to 1us
adc.delay = round((gx.rise_time + gx.flat_time / 2 - time_to_center) * 1e6) * 1e-6
# This rounding actually makes the sampling points on odd and even readouts to appear misaligned. However, on the real
# hardware this misalignment is much stronger anyways due to the gradient delays

# Split the blip into two halves and produnce a combined synthetic gradient
gy_parts = pp.split_gradient_at(grad=gy, time_point=blip_duration / 2, system=system)
gy_blipup, gy_blipdown, _ = pp.align(right=gy_parts[0], left=[gy_parts[1], gx])
gy_blipdownup = pp.add_gradients((gy_blipdown, gy_blipup), system=system)

# pe_enable support
gy_blipup.waveform = gy_blipup.waveform * pe_enable
gy_blipdown.waveform = gy_blipdown.waveform * pe_enable
gy_blipdownup.waveform = gy_blipdownup.waveform * pe_enable

# Phase encoding and partial Fourier
Ny_pre = round(part_fourier_factor * Ny / 2 - 1)  # PE steps prior to ky=0, excluding the central line
Ny_post = round(Ny / 2 + 1)  # PE lines after the k-space center including the central line
Ny_meas = Ny_pre + Ny_post

# Pre-phasing gradients
gx_pre = pp.make_trapezoid(channel='x', system=system, area=-gx.area / 2)
gy_pre = pp.make_trapezoid(channel='y', system=system, area=Ny_pre * delta_k)

gx_pre, gy_pre = pp.align(right=gx_pre, left=gy_pre)
# Relax the PE prephaser to reduce stimulation
gy_pre = pp.make_trapezoid('y', system=system, area=gy_pre.area, duration=pp.calc_duration(gx_pre, gy_pre))
gy_pre.amplitude = gy_pre.amplitude * pe_enable

# Calculate delay times
duration_to_center = (Ny_pre + 0.5) * pp.calc_duration(gx)
rf_center_incl_delay = rf.delay + pp.calc_rf_center(rf)[0]
rf180_center_incl_delay = rf180.delay + pp.calc_rf_center(rf180)[0]
delay_TE1 = math.ceil((TE / 2 - pp.calc_duration(rf,
                                                 gz) + rf_center_incl_delay - rf180_center_incl_delay) / system.grad_raster_time) * system.grad_raster_time
delay_TE2 = math.ceil((TE / 2 - pp.calc_duration(rf180,
                                                 gz180n) + rf180_center_incl_delay - duration_to_center) / system.grad_raster_time) * system.grad_raster_time
assert (delay_TE1 >= 0)
# Now we merge slice refocusing, TE delay and pre-phasers into a single block
delay_TE2 = delay_TE2 + pp.calc_duration(rf180, gz180n)
gx_pre.delay = 0
gx_pre.delay = delay_TE2 - pp.calc_duration(gx_pre)
assert (gx_pre.delay >= pp.calc_duration(rf180))  # gx_pre may not overlap with the RF
gy_pre.delay = pp.calc_duration(rf180)
assert (pp.calc_duration(gy_pre) <= pp.calc_duration(gx_pre))  # gyPre may not shift the timing

# ======
# CONSTRUCT SEQUENCE
# ======
# Define sequence blocks
for s in range(n_slices):
    seq.add_block(rf_fs, gz_fs)
    rf.freq_offset = gz.amplitude * slice_thickness * (s - (n_slices - 1) / 2)
    rf180.freq_offset = gz180.amplitude * slice_thickness * (s - (n_slices - 1) / 2)
    seq.add_block(rf, gz, trig)
    seq.add_block(pp.make_delay(delay_TE1))
    seq.add_block(rf180, gz180n, pp.make_delay(delay_TE2), gx_pre, gy_pre)
    for i in range(1, Ny_meas + 1):
        if i == 1:
            seq.add_block(gx, gy_blipup, adc)  # Read the first line of k-space with a single half-blip at the end
        elif i == Ny_meas:
            # Read the last line of k-space with a single half-blip at the beginning
            seq.add_block(gx, gy_blipdown, adc)
        else:
            # Read an intermediate line of k-space with a half-blip at the beginning and a half-blip at the end
            seq.add_block(gx, gy_blipdownup, adc)
        gx.amplitude = -gx.amplitude  # Reverse polarity of read gradient

ok, error_report = seq.check_timing()  # Check whether the timing of the sequence is correct
if ok:
    print('Timing check passed successfully')
else:
    print('Timing check failed. Error listing follows:')
    [print(e) for e in error_report]

# ======
# VISUALIZATION
# ======
seq.plot()

# Trajectory calculation and plotting
ktraj_adc, ktraj, t_excitation, t_refocusing, t_adc = seq.calculate_kspace()

time_axis = np.arange(1, ktraj.shape[1] + 1) * system.grad_raster_time
plt.figure()
plt.plot(time_axis, ktraj.T)  # Plot the entire k-space trajectory
plt.plot(t_adc, ktraj_adc[0], '.')  # Plot sampling points on the kx-axis
plt.figure()
plt.plot(ktraj[0], ktraj[1], 'b')  # 2D plot
plt.axis('equal')  # Enforce aspect ratio for the correct trajectory display
plt.plot(ktraj_adc[0], ktraj_adc[1], 'r.')  # Plot  sampling points
plt.show()

# Prepare the sequence output for the scanner
seq.set_definition('FOV', [fov, fov, slice_thickness])
seq.set_definition('Name', 'epi')

seq.write('epi_se_rs_pypulseq.seq')

# Very optional slow step, but useful for testing during development e.g. for the real TE, TR or for staying within
# slewrate limits
rep = seq.test_report()
print(rep)



from typing import Tuple

import setuptools


def _get_version() -> Tuple[int, int, int]:
    """
    Returns version of current PyPulseq release.

    Returns
    -------
    major, minor, revision : int
        Major, minor and revision numbers of current PyPulseq release.
    """
    with open('VERSION', 'r') as version_file:
        major, minor, revision = version_file.read().strip().split('.')
    return major, minor, revision


def _get_long_description() -> str:
    """
    Returns long description from `README.md` if possible, else 'Pulseq in Python'.

    Returns
    -------
    str
        Long description of PyPulseq project.
    """
    try:  # Unicode decode error on Windows
        with open("README.md", "r") as fh:
            long_description = fh.read()
    except:
        long_description = 'Pulseq in Python'
    return long_description


setuptools.setup(
    author="Keerthi Sravan Ravi",
    author_email="ks3621@columbia.edu",
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Operating System :: OS Independent",
    ],
    description="Pulseq in Python",
    include_package_data=True,
    install_requires=[
        'matplotlib>=3.3.4',
        'numpy>=1.19.5',
        'scipy>=1.5.4'
    ],
    license='License :: OSI Approved :: GNU Affero General Public License v3',
    long_description=_get_long_description(),
    long_description_content_type="text/markdown",
    name="pypulseq",
    packages=setuptools.find_packages(),
    # package_data for wheel distributions; MANIFEST.in for source distributions
    package_data={
        '': ['../VERSION'],
        'pypulseq.SAR': ['QGlobal.mat']
    },
    project_urls={
        'Documentation': 'https://pypulseq.readthedocs.io/en/latest/'
    },
    python_requires='>=3.6.3',
    url="https://github.com/imr-framework/pypulseq",
    version=".".join(_get_version()),
)



import math
from types import SimpleNamespace
from typing import Tuple

import numpy as np
from scipy.optimize import minimize

from pypulseq.make_extended_trapezoid import make_extended_trapezoid
from pypulseq.opts import Opts


def make_extended_trapezoid_area(channel: str, Gs: float, Ge: float, A: float,
                                 system: Opts) -> Tuple[SimpleNamespace, np.array, np.array]:
    """
    Makes shortest possible extended trapezoid with a given area.

    Parameters
    ----------
    channel : str
        Orientation of extended trapezoidal gradient event. Must be one of 'x', 'y' or 'z'.
    Gs : float
        Starting non-zero gradient value.
    Ge : float
        Ending non-zero gradient value.
    A : float
        Area of extended trapezoid.
    system: Opts
        System limits.

    Returns
    -------
    grad : SimpleNamespace
        Extended trapezoid event.
    times : numpy.ndarray
    amplitude : numpy.ndarray

    Raises
    ------
    ValueError

    """
    SR = system.max_slew * 0.99

    Tp = 0
    obj1 = lambda x: (A - __testGA(x, 0, SR, system.grad_raster_time, Gs, Ge)) ** 2
    res = minimize(fun=obj1, x0=0, method='Nelder-Mead')
    Gp, obj1val = *res.x, res.fun

    if obj1val > 1e-3 or abs(Gp) > system.max_grad:  # Search did not converge
        Gp = system.max_grad * np.sign(Gp)
        obj2 = lambda x: (A - __testGA(Gp, x, SR, system.grad_raster_time, Gs, Ge)) ** 2
        res2 = minimize(fun=obj2, x0=0, method='Nelder-Mead')
        T, obj2val = *res2.x, res2.fun
        assert obj2val < 1e-2

        Tp = math.ceil(T / system.grad_raster_time) * system.grad_raster_time

        # Fix the ramps
        Tru = math.ceil(abs(Gp - Gs) / SR / system.grad_raster_time) * system.grad_raster_time
        Trd = math.ceil(abs(Gp - Ge) / SR / system.grad_raster_time) * system.grad_raster_time
        obj3 = lambda x: (A - __testGA1(x, Tru, Tp, Trd, Gs, Ge)) ** 2

        res = minimize(fun=obj3, x0=Gp, method='Nelder-Mead')
        Gp, obj3val = *res.x, res.fun
        assert obj3val < 1e-3

    if Tp > 0:
        times = np.cumsum([0, Tru, Tp, Trd])
        amplitudes = [Gs, Gp, Gp, Ge]
    else:
        Tru = math.ceil(abs(Gp - Gs) / SR / system.grad_raster_time) * system.grad_raster_time
        Trd = math.ceil(abs(Gp - Ge) / SR / system.grad_raster_time) * system.grad_raster_time

        if Trd > 0:
            if Tru > 0:
                times = np.cumsum([0, Tru, Trd])
                amplitudes = np.array([Gs, Gp, Ge])
            else:
                times = np.cumsum([0, Trd])
                amplitudes = np.array([Gs, Ge])
        else:
            times = np.cumsum([0, Tru])
            amplitudes = np.array([Gs, Ge])

    grad = make_extended_trapezoid(channel=channel, system=system, times=times, amplitudes=amplitudes)

    return grad, times, amplitudes


def __testGA(Gp, Tp, SR, dT, Gs, Ge):
    Tru = math.ceil(abs(Gp - Gs) / SR / dT) * dT
    Trd = math.ceil(abs(Gp - Ge) / SR / dT) * dT
    ga = __testGA1(Gp, Tru, Tp, Trd, Gs, Ge)
    return ga


def __testGA1(Gp, Tru, Tp, Trd, Gs, Ge):
    return 0.5 * Tru * (Gp + Gs) + Gp * Tp + 0.5 * (Gp + Ge) * Trd
